import asyncio
from collections.abc import Callable

import pydantic
import pytest

from semlib.compare import _DEFAULT_TEMPLATE, _DEFAULT_TEMPLATE_BY
from semlib.sort import Sort, sort, sort_sync
from semlib.sort.algorithm import BordaCount, QuickSort
from tests.conftest import LLMMocker


def test_sort(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    measurements: list[str] = ["1 mile", "1.4 kilometers", "3 inches", "3 centimeters", "1 foot"]

    mocker = llm_mocker(
        {
            _DEFAULT_TEMPLATE.format(a="1 mile", b="1.4 kilometers"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a="1 mile", b="1 foot"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a="1 mile", b="3 inches"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a="1 mile", b="3 centimeters"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a="1.4 kilometers", b="1 foot"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a="1.4 kilometers", b="3 inches"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a="1.4 kilometers", b="3 centimeters"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a="1 foot", b="3 inches"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a="1 foot", b="3 centimeters"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a="3 inches", b="3 centimeters"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(b="1 mile", a="1.4 kilometers"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(b="1 mile", a="1 foot"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(b="1 mile", a="3 inches"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(b="1 mile", a="3 centimeters"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(b="1.4 kilometers", a="1 foot"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(b="1.4 kilometers", a="3 inches"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(b="1.4 kilometers", a="3 centimeters"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(b="1 foot", a="3 inches"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(b="1 foot", a="3 centimeters"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(b="3 inches", a="3 centimeters"): '{"choice": "B"}',
        }
    )

    sorter = Sort()
    with mocker.patch_llm():
        sort_measurements: list[str] = asyncio.run(sorter.sort(measurements, algorithm=QuickSort(randomized=False)))
    assert sort_measurements == ["3 centimeters", "3 inches", "1 foot", "1.4 kilometers", "1 mile"]
    assert sorter.total_cost() == 9.0

    with mocker.patch_llm():
        sort_measurements = sort_sync(measurements, algorithm=BordaCount())
    assert sort_measurements == ["3 centimeters", "3 inches", "1 foot", "1.4 kilometers", "1 mile"]


@pytest.mark.asyncio
async def test_sort_equal(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            _DEFAULT_TEMPLATE.format(a="12 inches", b="1 meter"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(a="1 foot", b="1 meter"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(a="1 foot", b="12 inches"): '{"choice": "neither"}',
        }
    )

    items = ["1 meter", "12 inches", "1 foot"]
    with mocker.patch_llm():
        result = await sort(items, algorithm=QuickSort(randomized=False), task="choose_greater_or_abstain")
    assert result == ["12 inches", "1 foot", "1 meter"]


def test_sort_by(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            _DEFAULT_TEMPLATE_BY.format(criteria="wavelength", a="red", b="green"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE_BY.format(criteria="wavelength", a="red", b="blue"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE_BY.format(criteria="wavelength", a="green", b="blue"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE_BY.format(criteria="wavelength", a="green", b="red"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE_BY.format(criteria="wavelength", a="blue", b="red"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE_BY.format(criteria="wavelength", a="blue", b="green"): '{"choice": "B"}',
        }
    )
    colors = ["red", "green", "blue"]
    with mocker.patch_llm():
        result = sort_sync(colors, by="wavelength")
    assert result == ["blue", "green", "red"]


@pytest.mark.asyncio
async def test_sort_raises_quicksort(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            _DEFAULT_TEMPLATE.format(a="12 inches", b="1 meter"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(a="1 foot", b="1 meter"): '{"choice": "1 meter"}',  # invalid
            _DEFAULT_TEMPLATE.format(a="1 foot", b="12 inches"): '{"choice": "neither"}',
        }
    )

    items = ["1 meter", "12 inches", "1 foot"]
    with mocker.patch_llm(), pytest.raises(pydantic.ValidationError):
        await sort(items, algorithm=QuickSort(randomized=False))


@pytest.mark.asyncio
async def test_sort_raises_borda_count(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            _DEFAULT_TEMPLATE.format(a="12 inches", b="1 meter"): '{"choice": "invalid"}',
            _DEFAULT_TEMPLATE.format(a="1 meter", b="12 inches"): '{"choice": "invalid"}',
        }
    )

    items = ["1 meter", "12 inches"]
    with mocker.patch_llm(), pytest.raises(pydantic.ValidationError):
        await sort(items, algorithm=BordaCount(), max_concurrency=1)



import asyncio
import os
from typing import cast, overload

import litellm
from litellm.types.utils import Message
from pydantic import BaseModel
from pydantic_core import ValidationError  # noqa: F401 # used in docstrings

from semlib._internal.constants import DEFAULT_MODEL
from semlib._internal.util import parse_max_concurrency
from semlib.bare import Bare
from semlib.cache import QueryCache


class Base:
    def __init__(
        self, *, model: str | None = None, max_concurrency: int | None = None, cache: QueryCache | None = None
    ):
        """Initialize.

        Args:
            model: The language model to use for completions. If not specified, uses the value from the
                `SEMLIB_DEFAULT_MODEL` environment variable, or falls back to the default model (currently
                `"openai/gpt-4o"`). This is used as the `model` argument for
                [litellm](https://docs.litellm.ai/docs/providers) unless overridden in individual method calls.
            max_concurrency: Maximum number of concurrent API requests. If not specified, uses the value from the
                `SEMLIB_MAX_CONCURRENCY` environment variable, or defaults to 10 for most models, or 1 for Ollama
                models.
            cache: If provided, this is used to cache LLM responses to avoid redundant API calls.

        Raises:
            ValueError: If `max_concurrency` is provided but is not a positive integer.
        """
        self._model = model or os.getenv("SEMLIB_DEFAULT_MODEL") or DEFAULT_MODEL
        if max_concurrency is None and (env_max_concurrency := os.getenv("SEMLIB_MAX_CONCURRENCY")) is not None:
            try:
                max_concurrency = int(env_max_concurrency)
            except ValueError:
                msg = "SEMLIB_MAX_CONCURRENCY must be an integer"
                raise ValueError(msg) from None
        self._max_concurrency = parse_max_concurrency(max_concurrency, self._model)
        self._sem = asyncio.Semaphore(self._max_concurrency)
        self._pending_requests: set[bytes] = set()
        self._cond = asyncio.Condition()  # for pending requests deduplication
        self._cache = cache

        self._total_cost: float = 0.0

    def total_cost(self) -> float:
        """Get the total cost incurred so far for API calls made through this instance.

        Returns:
            The total cost in USD.
        """
        return self._total_cost

    def _add_cost(self, cost: float) -> None:
        self._total_cost += cost

    @overload
    async def _acompletion[T: BaseModel](
        self, *, messages: list[Message], return_type: type[T], model: str | None = None
    ) -> T: ...

    @overload
    async def _acompletion[T](
        self, *, messages: list[Message], return_type: Bare[T], model: str | None = None
    ) -> T: ...

    @overload
    async def _acompletion(
        self,
        *,
        messages: list[Message],
        return_type: None = None,
        model: str | None = None,
    ) -> str: ...

    async def _acompletion[T: BaseModel, U](
        self, *, messages: list[Message], model: str | None = None, return_type: type[T] | Bare[U] | None = None
    ) -> str | T | U:
        model = model if model is not None else self._model
        cache_key = (messages, return_type._model if isinstance(return_type, Bare) else return_type, model)  # noqa: SLF001

        # check cache / pending requests
        if self._cache is not None:
            hashed_key = self._cache._hash_key(cache_key)  # noqa: SLF001
            async with self._cond:
                # if there's already a request in flight for this key, wait for it to complete
                while hashed_key in self._pending_requests:
                    await self._cond.wait()
                # then, check the cache
                if (cached_response := self._cache._get(cache_key)) is not None:  # noqa: SLF001
                    # cache hit
                    if return_type is None:
                        return cached_response
                    if isinstance(return_type, Bare):
                        return return_type._extract(return_type._model.model_validate_json(cached_response))  # noqa: SLF001
                    return return_type.model_validate_json(cached_response)
                # if we've gotten to this point, we need to make a request; mark it as pending
                self._pending_requests.add(hashed_key)

        # make the request
        if return_type is None:
            kwargs = {}
        elif isinstance(return_type, Bare):
            kwargs = {"response_format": return_type._model}  # noqa: SLF001
        else:
            kwargs = {"response_format": return_type}
        try:
            async with self._sem:
                response = await litellm.acompletion(model=model, messages=messages, **kwargs)
            self._add_cost(litellm.completion_cost(response))  # type: ignore[attr-defined]
            content = cast(str, response.choices[0].message.content)
            if return_type is None:
                typed_response: str | T | U = content
            elif isinstance(return_type, Bare):
                typed_response = return_type._extract(return_type._model.model_validate_json(content))  # noqa: SLF001
            else:
                typed_response = return_type.model_validate_json(content)
            # cache after parsing, to ensure we don't cache a broken response
            if self._cache is not None:
                async with self._cond:
                    self._cache._set(cache_key, content)  # noqa: SLF001
            return typed_response
        finally:
            if self._cache is not None:
                async with self._cond:
                    self._pending_requests.remove(hashed_key)
                    self._cond.notify_all()

    @property
    def model(self) -> str:
        """Get the current model being used for completions.

        Returns:
            The model name as a string.
        """
        return self._model

    def clear_cache(self) -> None:
        """Clear the internal cache of LLM responses, if caching is enabled."""
        if self._cache is not None:
            self._cache.clear()

    @overload
    async def prompt[T: BaseModel](self, prompt: str, /, *, return_type: type[T], model: str | None = None) -> T: ...

    @overload
    async def prompt[T](self, prompt: str, /, *, return_type: Bare[T], model: str | None = None) -> T: ...

    @overload
    async def prompt(self, prompt: str, /, *, return_type: None = None, model: str | None = None) -> str: ...

    async def prompt[T: BaseModel, U](
        self, prompt: str, /, *, return_type: type[T] | Bare[U] | None = None, model: str | None = None
    ) -> str | T | U:
        """Send a prompt to the language model and get a response.

        This method sends a single user message to the language model and returns the response. The response can be
        returned as a raw string, parsed into a [Pydantic](https://pydantic.dev/) model, or extracted as a bare value
        using the [Bare][semlib.bare.Bare] marker.

        Args:
            prompt: The text prompt to send to the language model.
            return_type: If not specified, the response is returned as a raw string. If a Pydantic model class is
                provided, the response is parsed into an instance of that model. If a [Bare][semlib.bare.Bare] instance
                is provided, a single value of the specified type is extracted from the response.
            model: If specified, overrides the default model for this call.

        Returns:
            The language model's response in the format specified by return_type.

        Raises:
            ValidationError: If `return_type` is a Pydantic model or a Bare type and the response cannot be
                parsed into the specified type.

        Examples:
            Get raw string response:
            >>> await session.prompt("What is 2+2?")
            '2 + 2 equals 4.'

            Get structured value:
            >>> class Person(BaseModel):
            ...     name: str
            ...     age: int
            >>> await session.prompt("Who is Barack Obama?", return_type=Person)
            Person(name='Barack Obama', age=62)

            Get bare value:
            >>> await session.prompt("What is 2+2?", return_type=Bare(int))
            4
        """
        return await self._acompletion(
            messages=[Message(role="user", content=prompt)], return_type=return_type, model=model
        )



import datetime
from collections.abc import Callable

import pydantic
import pytest

from semlib.apply import apply, apply_sync
from semlib.bare import Bare
from tests.conftest import LLMMocker


def test_apply_sync(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    class Person(pydantic.BaseModel):
        name: str
        date_of_birth: datetime.date

    people = [
        Person(name="George Washington", date_of_birth=datetime.date(1732, 2, 22)),
        Person(name="John Adams", date_of_birth=datetime.date(1735, 10, 30)),
        Person(name="Thomas Jefferson", date_of_birth=datetime.date(1743, 4, 13)),
        Person(name="James Madison", date_of_birth=datetime.date(1751, 3, 16)),
        Person(name="James Monroe", date_of_birth=datetime.date(1758, 4, 28)),
    ]

    class Count(pydantic.BaseModel):
        count: int

    mocker = llm_mocker(
        {
            "Of the following people, how many have names that start with 'J'?\nGeorge Washington,John Adams,Thomas Jefferson,James Madison,James Monroe": '{"count": 3}',
        }
    )
    with mocker.patch_llm():
        result: Count = apply_sync(
            people,
            lambda people: f"Of the following people, how many have names that start with 'J'?\n{','.join([p.name for p in people])}",
            return_type=Count,
        )

    assert result == Count(count=3)


def test_apply_sync_item(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    class Person(pydantic.BaseModel):
        name: str
        date_of_birth: datetime.date

    person = Person(name="George Washington", date_of_birth=datetime.date(1732, 2, 22))

    class BirthYear(pydantic.BaseModel):
        year: int

    mocker = llm_mocker(
        {
            "What year was George Washington born?": '{"year": 1732}',
        }
    )
    with mocker.patch_llm():
        result: BirthYear = apply_sync(
            person,
            lambda p: f"What year was {p.name} born?",
            return_type=BirthYear,
        )

    assert result == BirthYear(year=1732)


def test_apply_sync_str(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "What is 7 + 12?": "19",
        }
    )
    with mocker.patch_llm():
        result: str = apply_sync(12, "What is 7 + {}?")

    assert result == "19"


@pytest.mark.asyncio
async def test_apply_async(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "What is 7 + 12?": "19",
        }
    )
    with mocker.patch_llm():
        result: str = await apply(
            (7, 12),
            lambda xs: "What is {} + {}?".format(*xs),
        )

    assert result == "19"


@pytest.mark.asyncio
async def test_apply_async_raises(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    class Count(pydantic.BaseModel):
        count: int

    mocker = llm_mocker(
        {
            "What is 7 + 12?": "nineteen",
        }
    )
    with mocker.patch_llm(), pytest.raises(pydantic.ValidationError):
        await apply(
            (7, 12),
            lambda xs: "What is {} + {}?".format(*xs),
            return_type=Count,
        )


def test_apply_sync_bare(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "What is 7 + 12?": '{"value": 19}',
        }
    )
    with mocker.patch_llm():
        result = apply_sync(12, "What is 7 + {}?", return_type=Bare(int))

    assert result == 19



import os
from collections.abc import Callable
from contextlib import ExitStack
from typing import Any
from unittest.mock import AsyncMock, patch

import pytest


class LLMMocker:
    def __init__(self, response_map: dict[str, str]):
        self.response_map = response_map

    def _create_mock_response(self, content: str) -> AsyncMock:
        response = AsyncMock()
        response.choices = [AsyncMock()]
        response.choices[0].message.content = content
        return response

    async def mock_acompletion(self, *_args: Any, **kwargs: Any) -> AsyncMock:
        messages = kwargs.get("messages", [])
        if messages and isinstance(messages, list) and len(messages) > 0:
            prompt = messages[0].get("content", "")

            if prompt in self.response_map:
                return self._create_mock_response(self.response_map[prompt])

            msg = f"no mock response found for prompt: {prompt}"
            raise ValueError(msg)
        msg = "no content in LLM query"
        raise ValueError(msg)

    def patch_llm(self) -> ExitStack:
        def context_manager() -> ExitStack:
            stack = ExitStack()
            if os.getenv("SEMLIB_TEST_REAL_LLM") != "1":
                stack.enter_context(patch("litellm.acompletion", side_effect=self.mock_acompletion))
            stack.enter_context(patch("litellm.completion_cost", return_value=1.0))
            return stack

        return context_manager()


@pytest.fixture
def llm_mocker() -> Callable[[dict[str, str]], LLMMocker]:
    def _create_mocker(response_map: dict[str, str]) -> LLMMocker:
        return LLMMocker(response_map)

    return _create_mocker



class Box[T]:
    """A container that holds a value of type `T`.

    This can be used to tag values, so that they can be distinguished from other values of the same underlying type.
    Such a tag can be useful in the context of methods like [reduce][semlib.reduce.Reduce.reduce], where you can use
    this marker to distinguish leaf nodes from internal nodes in an associative reduce.
    """

    def __init__(self, value: T) -> None:
        """Initialize a Box instance.

        Args:
            value: The value to be contained in the Box.
        """
        self._value = value

    @property
    def value(self) -> T:
        """Get the value contained in the Box.

        Returns:
            The value contained in the Box.
        """
        return self._value



import asyncio
from collections.abc import Callable, Iterable

from pydantic_core import ValidationError  # noqa: F401 # used in docstrings

from semlib._internal.base import Base
from semlib.filter import _DEFAULT_TEMPLATE, _Decision


class Find(Base):
    async def find[T](
        self,
        iterable: Iterable[T],
        /,
        *,
        by: str | None = None,
        to_str: Callable[[T], str] | None = None,
        template: str | Callable[[T], str] | None = None,
        negate: bool = False,
        model: str | None = None,
    ) -> T | None:
        """Find an item in an iterable based on a criteria.

        This method searches through the provided iterable and returns some item (not necessarily the first) that
        matches the specified criteria.

        Args:
            iterable: The collection of items to search.
            by: A criteria specifying a predicate to search by. If this is provided, `template` cannot be provided.
            to_str: If specified, used to convert items to string representation. Otherewise, uses `str()` on each item.
                If this is provided, a callable template cannot be provided.
            template: A custom prompt template for predicates. Must be either a string template with a single positional
                placeholder, or a callable that takes an item and returns a formatted string. If this is provided, `by`
                cannot be provided.
            negate: If `True`, find an item that does **not** match the criteria. If `False`, find an item that does
                match the criteria.
            model: If specified, overrides the default model for this call.

        Returns:
            An item from the iterable if it matches the criteria, or `None` if no such item is found.

        Raises:
            ValidationError: If parsing any LLM response fails.

        Examples:
            Basic find:
            >>> await session.find(["Tom Hanks", "Tom Cruise", "Tom Brady"], by="actor?")
            'Tom Cruise'  # nondeterministic, could also return "Tom Hanks"

            Custom template:
            >>> await session.find(
            ...     [(123, 321), (384, 483), (134, 431)],
            ...     template=lambda pair: f"Is {pair[0]} backwards {pair[1]}?",
            ...     negate=True,
            ... )
            (384, 483)
        """
        if template is None:
            if by is None:
                msg = "must specify either 'by' or 'template'"
                raise ValueError(msg)
        else:
            if callable(template) and to_str is not None:
                msg = "cannot provide 'to_str' when a template function is provided"
                raise ValueError(msg)
            if by is not None:
                msg = "cannot provide 'by' when a custom template is provided"
                raise ValueError(msg)

        to_str = to_str if to_str is not None else str

        if template is None:

            def map_template(item: T, /) -> str:
                return _DEFAULT_TEMPLATE.format(by=by or "", item=to_str(item))
        elif isinstance(template, str):

            def map_template(item: T, /) -> str:
                return template.format(to_str(item))
        else:
            # callable
            map_template = template

        model = model if model is not None else self._model

        async def fn(item: T) -> tuple[T, bool]:
            decision = await self.prompt(
                map_template(item),
                return_type=_Decision,
                model=model,
            )
            if negate:
                decision.decision = not decision.decision
            return item, decision.decision

        tasks: list[asyncio.Task[tuple[T, bool]]] = [asyncio.create_task(fn(item)) for item in iterable]
        try:
            for next_finished in asyncio.as_completed(tasks):
                item, decision = await next_finished
                if decision:
                    return item
            return None
        finally:
            for task in tasks:
                if not task.done():
                    task.cancel()
            await asyncio.wait(tasks)


async def find[T](
    iterable: Iterable[T],
    /,
    *,
    by: str | None = None,
    to_str: Callable[[T], str] | None = None,
    template: str | Callable[[T], str] | None = None,
    negate: bool = False,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> T | None:
    """Standalone version of [find][semlib.find.Find.find]."""
    finder = Find(model=model, max_concurrency=max_concurrency)
    result = await finder.find(iterable, by=by, to_str=to_str, template=template, negate=negate)
    # binding to intermediate variable coro to avoid mypy bug, see https://github.com/python/mypy/issues/19716 and
    # https://github.com/python/mypy/pull/19767 (fixed now, but not shipped yet)
    return result  # noqa: RET504


def find_sync[T](
    iterable: Iterable[T],
    /,
    *,
    by: str | None = None,
    to_str: Callable[[T], str] | None = None,
    template: str | Callable[[T], str] | None = None,
    negate: bool = False,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> T | None:
    """Standalone synchronous version of [find][semlib.find.Find.find]."""
    finder = Find(model=model, max_concurrency=max_concurrency)
    coro = finder.find(iterable, by=by, to_str=to_str, template=template, negate=negate)
    # binding to intermediate variable coro to avoid mypy bug, see https://github.com/python/mypy/issues/19716 and
    # https://github.com/python/mypy/pull/19767 (fixed now, but not shipped yet)
    return asyncio.run(coro)



from semlib.apply import Apply
from semlib.extrema import Extrema
from semlib.filter import Filter
from semlib.find import Find
from semlib.reduce import Reduce
from semlib.sort import Sort


# don't need to include
# - Map (is a subclass of Filter)
# - Compare (is a subclass of Sort)
class Session(Sort, Filter, Extrema, Find, Apply, Reduce):
    """A session provides a context for performing Semlib operations.

    Sessions provide defaults (e.g., default model), manage caching, implement concurrency control, and track costs.

    All of a Session's methods have analogous standalone functions (e.g., the [Session.map][semlib.map.Map.map] method
    has an analogous [map][semlib.map.map] function). It is recommended to use a Session for any non-trivial use of the
    semlib library.
    """



import asyncio
from collections.abc import Callable
from typing import overload

from pydantic import BaseModel
from pydantic_core import ValidationError  # noqa: F401 # used in docstrings

from semlib._internal.base import Base
from semlib.bare import Bare


class Apply(Base):
    @overload
    async def apply[T, U: BaseModel](
        self,
        item: T,
        /,
        template: str | Callable[[T], str],
        *,
        return_type: type[U],
        model: str | None = None,
    ) -> U: ...

    @overload
    async def apply[T, U](
        self,
        item: T,
        /,
        template: str | Callable[[T], str],
        *,
        return_type: Bare[U],
        model: str | None = None,
    ) -> U: ...

    @overload
    async def apply[T](
        self,
        item: T,
        /,
        template: str | Callable[[T], str],
        *,
        return_type: None = None,
        model: str | None = None,
    ) -> str: ...

    async def apply[T, U: BaseModel, V](
        self,
        item: T,
        /,
        template: str | Callable[[T], str],
        *,
        return_type: type[U] | Bare[V] | None = None,
        model: str | None = None,
    ) -> U | V | str:
        """Apply a language model prompt to a single item.

        This method formats a prompt template with the given item, sends it to the language model, and returns the
        response. The response can be returned as a raw string, parsed into a [Pydantic](https://pydantic.dev/) model,
        or extracted as a bare value using the [Bare][semlib.bare.Bare] marker.

        This method is a simple wrapper around [prompt][semlib._internal.base.Base.prompt].

        Args:
            item: The item to apply the `template` to.
            template: A template to format with the item. This can be either a string template with a single positional
                placeholder, or a callable that takes the item and returns a formatted string.
            return_type: If not specified, the response is returned as a raw string. If a Pydantic model class is
                provided, the response is parsed into an instance of that model. If a [Bare][semlib.bare.Bare] instance
                is provided, a single value of the specified type is extracted from the response.
            model: If specified, overrides the default model for this call.

        Returns:
            The language model's response in the format specified by return_type.

        Raises:
            ValidationError: If `return_type` is a Pydantic model or a Bare type and the response cannot be
                parsed into the specified type.

        Examples:
            Basic usage:
            >>> await session.apply(
            ...     [1, 2, 3, 4, 5],
            ...     template="What is the sum of these numbers: {}?",
            ...     return_type=Bare(int),
            ... )
            15
        """
        formatter = template.format if isinstance(template, str) else template

        model = model if model is not None else self._model

        return await self.prompt(formatter(item), return_type=return_type, model=model)


@overload
async def apply[T, U: BaseModel](
    item: T,
    /,
    template: str | Callable[[T], str],
    *,
    return_type: type[U],
    model: str | None = None,
) -> U: ...


@overload
async def apply[T, U](
    item: T,
    /,
    template: str | Callable[[T], str],
    *,
    return_type: Bare[U],
    model: str | None = None,
) -> U: ...


@overload
async def apply[T](
    item: T,
    /,
    template: str | Callable[[T], str],
    *,
    return_type: None = None,
    model: str | None = None,
) -> str: ...


async def apply[T, U: BaseModel, V](
    item: T,
    /,
    template: str | Callable[[T], str],
    *,
    return_type: type[U] | Bare[V] | None = None,
    model: str | None = None,
) -> U | V | str:
    """Standalone version of [apply][semlib.apply.Apply.apply]."""
    applier = Apply(model=model)
    return await applier.apply(item, template, return_type=return_type)


@overload
def apply_sync[T, U: BaseModel](
    item: T,
    /,
    template: str | Callable[[T], str],
    *,
    return_type: type[U],
    model: str | None = None,
) -> U: ...


@overload
def apply_sync[T, U](
    item: T,
    /,
    template: str | Callable[[T], str],
    *,
    return_type: Bare[U],
    model: str | None = None,
) -> U: ...


@overload
def apply_sync[T](
    item: T,
    /,
    template: str | Callable[[T], str],
    *,
    return_type: None = None,
    model: str | None = None,
) -> str: ...


def apply_sync[T, U: BaseModel, V](
    item: T,
    /,
    template: str | Callable[[T], str],
    *,
    return_type: type[U] | Bare[V] | None = None,
    model: str | None = None,
) -> U | V | str:
    """Standalone synchronous version of [apply][semlib.apply.Apply.apply]."""
    applier = Apply(model=model)
    return asyncio.run(applier.apply(item, template, return_type=return_type))  # type: ignore[return-value]



from collections.abc import Callable

import pydantic
import pytest

from semlib import Bare
from semlib.prompt import prompt, prompt_sync
from tests.conftest import LLMMocker


def test_prompt_sync(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "What is 1 + 1?": "2",
        }
    )
    with mocker.patch_llm():
        response = prompt_sync("What is 1 + 1?")
    assert response == "2"


@pytest.mark.asyncio
async def test_prompt_async(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    class Number(pydantic.BaseModel):
        value: int

    mocker = llm_mocker(
        {
            "What is 1 + 1?": '{"value": 2}',
        }
    )
    with mocker.patch_llm():
        response = await prompt("What is 1 + 1?", return_type=Number)
    assert response == Number(value=2)


@pytest.mark.asyncio
async def test_prompt_bare(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "Return the first 3 primes": '{"value": [2, 3, 5]}',
        }
    )
    b: Bare[list[int]] = Bare(list[int])
    with mocker.patch_llm():
        x: list[int] = await prompt("Return the first 3 primes", return_type=b)
    assert x == [2, 3, 5]


@pytest.mark.asyncio
async def test_prompt_bare_named(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "Return the list": '{"primes": [2, 3, 5]}',
        }
    )
    b: Bare[list[int]] = Bare(list[int], class_name="list_of_length_3", field_name="primes")
    with mocker.patch_llm():
        x: list[int] = await prompt("Return the list", return_type=b)
    assert x == [2, 3, 5]



import asyncio
from collections.abc import Callable, Iterable

from pydantic import BaseModel
from pydantic_core import ValidationError  # noqa: F401 # used in docstrings

from semlib.map import Map


class _Decision(BaseModel):
    decision: bool


_DEFAULT_TEMPLATE = """
Given a criteria and an item, determine whether the item meets the criteria.

<criteria>
{by}
</criteria>

<item>
{item}
</item>
""".strip()


class Filter(Map):
    async def filter[T](
        self,
        iterable: Iterable[T],
        /,
        *,
        by: str | None = None,
        to_str: Callable[[T], str] | None = None,
        template: str | Callable[[T], str] | None = None,
        negate: bool = False,
        model: str | None = None,
    ) -> list[T]:
        """Filter an iterable based on a criteria.

        This method is analogous to Python's built-in
        [`filter`](https://docs.python.org/3/library/functions.html#filter) function.

        Args:
            iterable: The collection of items to filter.
            by: A criteria specifying a predicate to filter by. If this is provided, `template` cannot be provided.
            to_str: If specified, used to convert items to string representation. Otherewise, uses `str()` on each item.
                If this is provided, a callable template cannot be provided.
            template: A custom prompt template for predicates. Must be either a string template with a single positional
                placeholder, or a callable that takes an item and returns a formatted string. If this is provided, `by`
                cannot be provided.
            negate: If `True`, keep items that do **not** match the criteria. If `False`, keep items that match the
                criteria.
            model: If specified, overrides the default model for this call.

        Returns:
            A new list containing items from the iterable if they match the criteria.

        Raises:
            ValidationError: If parsing any LLM response fails.

        Examples:
            Basic filter:
            >>> await session.filter(["Tom Hanks", "Tom Cruise", "Tom Brady"], by="actor?")
            ['Tom Hanks', 'Tom Cruise']

            Custom template:
            >>> await session.filter(
            ...     [(123, 321), (384, 483), (134, 431)],
            ...     template=lambda pair: f"Is {pair[0]} backwards {pair[1]}?",
            ...     negate=True,
            ... )
            [(384, 483)]
        """
        if template is None:
            if by is None:
                msg = "must specify either 'by' or 'template'"
                raise ValueError(msg)
        else:
            if callable(template) and to_str is not None:
                msg = "cannot provide 'to_str' when a template function is provided"
                raise ValueError(msg)
            if by is not None:
                msg = "cannot provide 'by' when a custom template is provided"
                raise ValueError(msg)

        to_str = to_str if to_str is not None else str

        if template is None:

            def map_template(item: T, /) -> str:
                return _DEFAULT_TEMPLATE.format(by=by or "", item=to_str(item))
        elif isinstance(template, str):

            def map_template(item: T, /) -> str:
                return template.format(to_str(item))
        else:
            # callable
            map_template = template

        decisions = await self.map(iterable, map_template, return_type=_Decision, model=model)
        return [
            item
            for item, decision in zip(iterable, decisions, strict=False)
            if ((not decision.decision) if negate else decision.decision)
        ]


async def filter[T](  # noqa: A001
    iterable: Iterable[T],
    /,
    *,
    by: str | None = None,
    to_str: Callable[[T], str] | None = None,
    template: str | Callable[[T], str] | None = None,
    negate: bool = False,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> list[T]:
    """Standalone version of [filter][semlib.filter.Filter.filter]."""
    filterer = Filter(model=model, max_concurrency=max_concurrency)
    return await filterer.filter(iterable, by=by, to_str=to_str, template=template, negate=negate)


def filter_sync[T](
    iterable: Iterable[T],
    /,
    *,
    by: str | None = None,
    to_str: Callable[[T], str] | None = None,
    template: str | Callable[[T], str] | None = None,
    negate: bool = False,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> list[T]:
    """Standalone synchronous version of [filter][semlib.filter.Filter.filter]."""
    filterer = Filter(model=model, max_concurrency=max_concurrency)
    return asyncio.run(filterer.filter(iterable, by=by, to_str=to_str, template=template, negate=negate))



import pathlib
import sqlite3

import pytest
from litellm.types.utils import Message
from pydantic import BaseModel

from semlib.cache import _VERSION, _VERSION_KEY, InMemoryCache, OnDiskCache, QueryCache

# ruff: noqa: SLF001


def test_in_memory_cache() -> None:
    cache: QueryCache = InMemoryCache()
    key: tuple[list[Message], None, str] = ([], None, "gpt-4")
    assert cache._get(key) is None
    cache._set(key, "value1")
    assert cache._get(key) == "value1"
    cache._set(key, "value2")
    assert cache._get(key) == "value2"
    cache.clear()
    assert cache._get(key) is None


def test_on_disk_cache(tmp_path: pathlib.Path) -> None:
    cache_path = tmp_path / "cache.db"
    cache: QueryCache = OnDiskCache(str(cache_path))
    key: tuple[list[Message], None, str] = ([], None, "gpt-4")
    assert cache._get(key) is None
    cache._set(key, "value1")
    assert cache._get(key) == "value1"
    cache._set(key, "value2")
    assert cache._get(key) == "value2"

    # re-open and check persistence
    cache = OnDiskCache(str(cache_path))
    assert cache._get(key) == "value2"

    # check version mismatch
    conn = sqlite3.connect(str(cache_path), autocommit=True)
    conn.execute(f"UPDATE metadata SET value = '{int(_VERSION)+1}' WHERE key = '{_VERSION_KEY}'")
    conn.close()
    with pytest.raises(ValueError, match="cache version mismatch"):
        OnDiskCache(str(cache_path))


@pytest.mark.parametrize("cache_type", ["in_memory", "on_disk"])
def test_cache_interface(cache_type: str, tmp_path: pathlib.Path) -> None:
    if cache_type == "on_disk":
        cache_path = tmp_path / "cache.db"
        cache: QueryCache = OnDiskCache(str(cache_path))
    else:
        cache = InMemoryCache()

    # test that different messages/return types/models are cached separately
    class Foo(BaseModel):
        bar: int

    key1: tuple[list[Message], None, str] = ([Message(role="user", content="Hello")], None, "gpt-4.1")
    key2: tuple[list[Message], None, str] = ([Message(role="user", content="Hello")], None, "gpt-4o")  # different model
    key3: tuple[list[Message], None, str] = (
        [Message(role="user", content="Hello"), Message(role="user", content="world")],
        None,
        "gpt-4",
    )  # different content
    key4: tuple[list[Message], type[Foo], str] = (
        [Message(role="user", content="Hello")],
        Foo,
        "gpt-4o",
    )  # different return type

    assert len(cache) == 0

    cache._set(key1, "value1")
    assert cache._get(key1) == "value1"
    assert cache._get(key2) is None
    assert cache._get(key3) is None
    assert cache._get(key4) is None

    cache._set(key2, "value2")
    assert cache._get(key1) == "value1"
    assert cache._get(key2) == "value2"
    assert cache._get(key3) is None
    assert cache._get(key4) is None

    cache._set(key3, "value3")
    assert cache._get(key1) == "value1"
    assert cache._get(key2) == "value2"
    assert cache._get(key3) == "value3"
    assert cache._get(key4) is None

    cache._set(key4, "value4")
    assert cache._get(key1) == "value1"
    assert cache._get(key2) == "value2"
    assert cache._get(key3) == "value3"
    assert cache._get(key4) == "value4"

    assert len(cache) == 4
    cache.clear()
    assert len(cache) == 0



from typing import Any, cast

import pydantic


class Bare[T]:
    """A marker to indicate that a function should return a bare value of type `T`.

    This can be passed to the `return_type` parameter of functions like [prompt][semlib._internal.base.Base.prompt]. For
    situations where you want to extract a single value of a given base type (like `int` or `list[float]`), this is more
    convenient than the alternative of defining a Pydantic model with a single field for the purpose of extracting that
    value.

    Examples:
        Extract a bare value using prompt:
        >>> await session.prompt("What is 2+2?", return_type=Bare(int))
        4

        Influence model output using `class_name` and `field_name`:
        >>> await session.prompt(
        ...     "Give me a list",
        ...     return_type=Bare(
        ...         list[int], class_name="list_of_three_values", field_name="primes"
        ...     ),
        ... )
        [3, 7, 11]
    """

    def __init__(self, typ: type[T], /, class_name: str | None = None, field_name: str | None = None):
        """Initialize a Bare instance.

        Args:
            typ: The type of the bare value to extract.
            class_name: Name for a dynamically created Pydantic model class. If not provided, defaults to
                the name of `typ`. This name is visible to the LLM and may affect model output.
            field_name: Name for the field in the dynamically created Pydantic model that holds the bare value.
                If not provided, defaults to "value". This name is visible to the LLM and may affect model output.
        """
        self._typ = typ
        self._class_name = class_name if class_name is not None else typ.__name__
        self._field_name = field_name if field_name is not None else "value"
        field_definitions: Any = {self._field_name: (self._typ, ...)}
        self._model: type[pydantic.BaseModel] = pydantic.create_model(self._class_name, **field_definitions)

    def _extract(self, obj: Any) -> T:
        if isinstance(obj, self._model):
            return cast(T, getattr(obj, self._field_name))
        msg = f"expected instance of {self._model.__name__}, got {type(obj).__name__}"
        raise TypeError(msg)



from collections.abc import Callable

import pydantic
import pytest

from semlib.compare import _DEFAULT_TEMPLATE, Order, Task, compare, compare_sync
from tests.conftest import LLMMocker


def test_compare(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            _DEFAULT_TEMPLATE.format(a=1, b=2): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(a=2, b=1): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a=1, b=1): '{"choice": "neither"}',
        }
    )

    with mocker.patch_llm():
        assert compare_sync(2, 1) == Order.GREATER
        assert compare_sync(1, 2, task="choose_greater_or_abstain") == Order.LESS
        assert compare_sync(1, 1, task="choose_greater_or_abstain") == Order.NEITHER


def test_compare_choose_lesser(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    template = "Which is smaller, (a) {} or (b) {}?"
    mocker = llm_mocker(
        {
            template.format(2, 1): '{"choice": "B"}',
            template.format(1, 2): '{"choice": "A"}',
            template.format(1, 1): '{"choice": "neither"}',
            template.format(3, 9): '{"choice": "neither"}',
        }
    )

    with mocker.patch_llm():
        assert compare_sync(2, 1, template=template, task=Task.CHOOSE_LESSER) == Order.GREATER
        assert compare_sync(2, 1, template=template, task=Task.CHOOSE_LESSER_OR_ABSTAIN) == Order.GREATER
        assert compare_sync(1, 2, template=template, task=Task.CHOOSE_LESSER) == Order.LESS
        assert compare_sync(1, 2, template=template, task=Task.CHOOSE_LESSER_OR_ABSTAIN) == Order.LESS
        assert compare_sync(1, 1, template=template, task=Task.CHOOSE_LESSER_OR_ABSTAIN) == Order.NEITHER
        assert compare_sync(3, 9, template=template, task=Task.CHOOSE_LESSER_OR_ABSTAIN) == Order.NEITHER


def test_compare_compare(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "Consider the numbers (a) 1 and (b) 2. Is (a) greater than (b), or (a) less than (b)?": '{"order": "less"}',
            "Consider the numbers (a) 2 and (b) 1. Is (a) greater than (b), or (a) less than (b)?": '{"order": "greater"}',
            "Consider the numbers (a) 1 and (b) 1. Is (a) greater than (b), (a) less than (b), or neither?": '{"order": "neither"}',
        }
    )

    with mocker.patch_llm():
        assert (
            compare_sync(
                1,
                2,
                template="Consider the numbers (a) {} and (b) {}. Is (a) greater than (b), or (a) less than (b)?",
                task=Task.COMPARE,
            )
            == Order.LESS
        )
        assert (
            compare_sync(
                2,
                1,
                template="Consider the numbers (a) {} and (b) {}. Is (a) greater than (b), or (a) less than (b)?",
                task=Task.COMPARE,
            )
            == Order.GREATER
        )
        assert (
            compare_sync(
                1,
                1,
                template="Consider the numbers (a) {} and (b) {}. Is (a) greater than (b), (a) less than (b), or neither?",
                task=Task.COMPARE_OR_ABSTAIN,
            )
            == Order.NEITHER
        )


@pytest.mark.asyncio
async def test_compare_callable(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    def template(a: int, b: int) -> str:
        return f"Which is greater, (a) {a} or (b) {b}?"

    mocker = llm_mocker(
        {
            template(1, 2): '{"choice": "B"}',
        }
    )

    with mocker.patch_llm():
        result = await compare(1, 2, template=template)
    assert result == Order.LESS


def test_compare_bad_args() -> None:
    with pytest.raises(ValueError, match="if 'task' is not"):
        compare_sync(1, 2, task=Task.CHOOSE_LESSER)
    with pytest.raises(ValueError, match="cannot provide"):
        compare_sync(1, 2, template=lambda a, b: f"which is bigger, (a) {a} or (b) {b}?", to_str=str)
    with pytest.raises(ValueError, match="cannot provide"):
        compare_sync("car", "house", template="which is bigger, (a) {} or (b) {}?", by="weight")


def test_compare_raises(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            _DEFAULT_TEMPLATE.format(a=1, b=2): "not a json",
        }
    )

    with mocker.patch_llm(), pytest.raises(pydantic.ValidationError):
        compare_sync(1, 2)



from collections.abc import Callable

import pydantic
import pytest

from semlib.compare import _DEFAULT_TEMPLATE, _DEFAULT_TEMPLATE_BY
from semlib.extrema import max, max_sync, min, min_sync
from tests.conftest import LLMMocker


def test_min_max(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            _DEFAULT_TEMPLATE.format(a="1", b="3"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(a="3", b="2"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a="1", b="2"): '{"choice": "B"}',
        }
    )

    with mocker.patch_llm():
        biggest = max_sync([1, 3, 2])
        smallest = min_sync([1, 3, 2])

    assert biggest == 3
    assert smallest == 1


@pytest.mark.asyncio
async def test_min_max_async(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            _DEFAULT_TEMPLATE.format(a="1", b="3"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(a="3", b="2"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a="1", b="2"): '{"choice": "B"}',
        }
    )

    with mocker.patch_llm():
        biggest = await max([1, 3, 2])
        smallest = await min([1, 3, 2])

    assert biggest == 3
    assert smallest == 1


def test_equal(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    class Person(pydantic.BaseModel):
        name: str
        age: int

    people = [Person(name="Alice", age=30), Person(name="Bob", age=30), Person(name="Charlie", age=25)]
    mocker = llm_mocker(
        {
            _DEFAULT_TEMPLATE.format(a="30", b="30"): '{"choice": "neither"}',
            _DEFAULT_TEMPLATE.format(a="30", b="25"): '{"choice": "A"}',
        }
    )

    with mocker.patch_llm():
        oldest: Person = max_sync(people, to_str=lambda p: str(p.age), task="choose_greater_or_abstain")

    assert oldest == people[0]  # left-biased


def test_min_max_callable(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    class University(pydantic.BaseModel):
        name: str
        year_founded: int

    universities = [
        University(name="Harvard", year_founded=1636),
        University(name="Stanford", year_founded=1885),
        University(name="MIT", year_founded=1861),
    ]

    mocker = llm_mocker(
        {
            _DEFAULT_TEMPLATE_BY.format(a="Harvard", b="Stanford", criteria="endowment"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE_BY.format(a="Stanford", b="MIT", criteria="endowment"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE_BY.format(a="Harvard", b="MIT", criteria="endowment"): '{"choice": "A"}',
        }
    )

    with mocker.patch_llm():
        biggest: University = max_sync(universities, by="endowment", to_str=lambda u: u.name)
        smallest: University = min_sync(universities, by="endowment", to_str=lambda u: u.name)

    assert biggest == universities[0]  # Harvard
    assert smallest == universities[2]  # MIT


def test_min_max_empty() -> None:
    with pytest.raises(ValueError, match="iterable argument is empty"):
        max_sync([])

    with pytest.raises(ValueError, match="iterable argument is empty"):
        min_sync([])


def test_min_raises(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    class Item(pydantic.BaseModel):
        name: str
        value: int

    items = [Item(name="A", value=10), Item(name="B", value=20)]

    mocker = llm_mocker(
        {
            _DEFAULT_TEMPLATE.format(a="10", b="20"): "invalid json",
        }
    )

    with mocker.patch_llm(), pytest.raises(pydantic.ValidationError):
        min_sync(items, to_str=lambda item: str(item.value))



from collections.abc import Callable

import pydantic
import pytest

from semlib.filter import _DEFAULT_TEMPLATE, filter, filter_sync
from tests.conftest import LLMMocker


@pytest.mark.parametrize("negate", [True, False])
def test_filter(llm_mocker: Callable[[dict[str, str]], LLMMocker], negate: bool) -> None:  # noqa: FBT001
    people = ["Barack Obama", "Miley Cyrus", "Jeff Dean", "Jeff Bezos", "Bill Clinton", "Elon Musk"]
    mocker = llm_mocker(
        {
            _DEFAULT_TEMPLATE.format(item="Barack Obama", by="president of the United States"): '{"decision": true}',
            _DEFAULT_TEMPLATE.format(item="Miley Cyrus", by="president of the United States"): '{"decision": false}',
            _DEFAULT_TEMPLATE.format(item="Jeff Dean", by="president of the United States"): '{"decision": false}',
            _DEFAULT_TEMPLATE.format(item="Jeff Bezos", by="president of the United States"): '{"decision": false}',
            _DEFAULT_TEMPLATE.format(item="Bill Clinton", by="president of the United States"): '{"decision": true}',
            _DEFAULT_TEMPLATE.format(item="Elon Musk", by="president of the United States"): '{"decision": false}',
        }
    )
    with mocker.patch_llm():
        result = filter_sync(people, by="president of the United States", negate=negate)

    if negate:
        assert result == ["Miley Cyrus", "Jeff Dean", "Jeff Bezos", "Elon Musk"]
    else:
        assert result == ["Barack Obama", "Bill Clinton"]


def test_filter_template_str(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    people = ["Barack Obama", "Miley Cyrus", "Jeff Dean", "Jeff Bezos", "Bill Clinton", "Elon Musk"]
    mocker = llm_mocker(
        {
            "Is Barack Obama a famous computer scientist?": '{"decision": false}',
            "Is Miley Cyrus a famous computer scientist?": '{"decision": false}',
            "Is Jeff Dean a famous computer scientist?": '{"decision": true}',
            "Is Jeff Bezos a famous computer scientist?": '{"decision": false}',
            "Is Bill Clinton a famous computer scientist?": '{"decision": false}',
            "Is Elon Musk a famous computer scientist?": '{"decision": false}',
        }
    )
    with mocker.patch_llm():
        result = filter_sync(people, template="Is {} a famous computer scientist?")
    assert result == ["Jeff Dean"]


@pytest.mark.asyncio
async def test_filter_callable(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    pairs = [
        (123, 321),
        (1838, 1883),
        (33, 44),
        (12, 21),
    ]
    mocker = llm_mocker(
        {
            "Is 123 the reverse of 321?": '{"decision": true}',
            "Is 12 the reverse of 21?": '{"decision": true}',
            "Is 1838 the reverse of 1883?": '{"decision": false}',
            "Is 33 the reverse of 44?": '{"decision": false}',
        }
    )
    with mocker.patch_llm():
        result = await filter(pairs, template=lambda pair: f"Is {pair[0]} the reverse of {pair[1]}?")
    assert result == [(123, 321), (12, 21)]


def test_filter_bad_args() -> None:
    with pytest.raises(ValueError, match="must specify either"):
        filter_sync([1, 2, 3])
    with pytest.raises(ValueError, match="cannot provide"):
        filter_sync([1, 2, 3], template=lambda x: f"Is {x} odd?", to_str=str)
    with pytest.raises(ValueError, match="cannot provide"):
        filter_sync([1, 2, 3], by="odd", template="Is {} odd?")


def test_filter_raises(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    people = ["Barack Obama", "Miley Cyrus", "Jeff Dean"]
    mocker = llm_mocker(
        {
            _DEFAULT_TEMPLATE.format(item="Barack Obama", by="president of the United States"): "not json",
            _DEFAULT_TEMPLATE.format(item="Miley Cyrus", by="president of the United States"): '{"decision": false}',
            _DEFAULT_TEMPLATE.format(item="Jeff Dean", by="president of the United States"): '{"decision": false}',
        }
    )
    with mocker.patch_llm(), pytest.raises(pydantic.ValidationError):
        filter_sync(people, by="president of the United States")



from abc import ABC, abstractmethod
from collections.abc import Callable, Coroutine, Iterable

from semlib.compare import Order


class Algorithm(ABC):
    """Abstract base class for sorting algorithms.

    Sorting algorithms can be used with [sort][semlib.sort.Sort.sort]."""

    def __init__(self) -> None:
        """Initialize."""

    @abstractmethod
    async def _sort[T](
        self,
        iterable: Iterable[T],
        /,
        *,
        reverse: bool = False,
        comparator: Callable[[T, T], Coroutine[None, None, Order]],
        max_concurrency: int,
    ) -> list[T]: ...



from collections.abc import Callable

import litellm
import pytest

from tests.conftest import LLMMocker


@pytest.mark.asyncio
async def test_exact_match_returns_expected_response(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    response_map = {"Hello world": "Hi there", "What is 2+2?": "4"}
    mocker = llm_mocker(response_map)

    with mocker.patch_llm():
        response = await litellm.acompletion(
            model="gpt-3.5-turbo", messages=[{"role": "user", "content": "Hello world"}]
        )
        assert response.choices[0].message.content == "Hi there"


@pytest.mark.asyncio
async def test_no_match_raises_error(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    response_map = {"Hello world": "Hi there"}
    mocker = llm_mocker(response_map)

    with mocker.patch_llm(), pytest.raises(ValueError, match="no mock response found for prompt: Goodbye world"):
        await litellm.acompletion(model="gpt-3.5-turbo", messages=[{"role": "user", "content": "Goodbye world"}])


@pytest.mark.asyncio
async def test_partial_match_does_not_work(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    response_map = {"Hello": "Hi there"}
    mocker = llm_mocker(response_map)

    with mocker.patch_llm(), pytest.raises(ValueError, match="no mock response found for prompt: Hello world"):
        await litellm.acompletion(model="gpt-3.5-turbo", messages=[{"role": "user", "content": "Hello world"}])



from semlib import apply, compare, extrema, filter, find, map, prompt, sort
from semlib.bare import Bare
from semlib.box import Box
from semlib.cache import InMemoryCache, OnDiskCache, QueryCache
from semlib.session import Session

__all__ = [
    "Bare",
    "Box",
    "InMemoryCache",
    "OnDiskCache",
    "QueryCache",
    "Session",
    "apply",
    "compare",
    "extrema",
    "filter",
    "find",
    "map",
    "prompt",
    "sort",
]



import asyncio
from collections.abc import Callable
from enum import Enum
from typing import cast

from pydantic import BaseModel
from pydantic_core import ValidationError  # noqa: F401 # used in docstrings

from semlib._internal.base import Base

_DEFAULT_TEMPLATE = """
Given two items, determine which item is greater.

<item A>
{a}
</item A>

<item B>
{b}
</item B>
""".strip()

_DEFAULT_TEMPLATE_BY = """
Given a criteria and two items, determine which item is greater according to the criteria.

<criteria>
{criteria}
</criteria>

<item A>
{a}
</item A>

<item B>
{b}
</item B>
""".strip()


class Task(str, Enum):
    """Comparison task to perform.

    Intended to be passed to [compare][semlib.compare.Compare.compare] and similar methods, this specifies how the LLM
    should compare two items.
    """

    COMPARE = "compare"
    """Ask the model to compare two items and determine their relative order.

    The model must choose either `"less"` or `"greater"`.
    """

    COMPARE_OR_ABSTAIN = "compare_or_abstain"
    """Ask the model to compare two items and determine their relative order, or abstain if unsure.

    The model must choose `"less"`, `"greater"`, or `"neither"`.
    """

    CHOOSE_GREATER = "choose_greater"
    """Ask the model to choose which of the two items (a) or (b) is greater.

    The model must choose either `"A"` or `"B"`.
    """

    CHOOSE_GREATER_OR_ABSTAIN = "choose_greater_or_abstain"
    """Ask the model to choose which of the two items (a) or (b) is greater, or abstain if unsure.

    The model must choose `"A"`, `"B"`, or `"neither"`.
    """

    CHOOSE_LESSER = "choose_lesser"
    """Ask the model to choose which of the two items (a) or (b) is lesser.

    The model must choose either `"A"` or `"B"`.
    """

    CHOOSE_LESSER_OR_ABSTAIN = "choose_lesser_or_abstain"
    """Ask the model to choose which of the two items (a) or (b) is lesser, or abstain if unsure.

    The model must choose `"A"`, `"B"`, or `"neither"`.
    """


class Order(str, Enum):
    """Result of a comparison."""

    LESS = "less"
    """Item A is less than Item B."""

    GREATER = "greater"
    """Item A is greater than Item B."""

    NEITHER = "neither"
    """Items A and B are equivalent, or the model abstained from choosing."""


class _StrictOrder(str, Enum):
    LESS = "less"
    GREATER = "greater"

    def to_order(self) -> Order:
        match self:
            case _StrictOrder.LESS:
                return Order.LESS
            case _StrictOrder.GREATER:
                return Order.GREATER


class _CompareResult(BaseModel):
    order: Order


class _StrictCompareResult(BaseModel):
    order: _StrictOrder


class _Choice(str, Enum):
    A = "A"
    B = "B"
    NEITHER = "neither"


class _StrictChoice(str, Enum):
    A = "A"
    B = "B"


class _ChooseResult(BaseModel):
    choice: _Choice


class _StrictChooseResult(BaseModel):
    choice: _StrictChoice


_RETURN_TYPE_BY_TASK: dict[Task, type[BaseModel]] = {
    Task.COMPARE: _StrictCompareResult,
    Task.COMPARE_OR_ABSTAIN: _CompareResult,
    Task.CHOOSE_GREATER: _StrictChooseResult,
    Task.CHOOSE_GREATER_OR_ABSTAIN: _ChooseResult,
    Task.CHOOSE_LESSER: _StrictChooseResult,
    Task.CHOOSE_LESSER_OR_ABSTAIN: _ChooseResult,
}


class Compare(Base):
    async def compare[T](
        self,
        a: T,
        b: T,
        /,
        *,
        by: str | None = None,
        to_str: Callable[[T], str] | None = None,
        template: str | Callable[[T, T], str] | None = None,
        task: Task | str | None = None,
        model: str | None = None,
    ) -> Order:
        """Compare two items.

        This method uses a language model to compare two items and determine the relative ordering of the two items.
        The comparison can be customized by specifying either a criteria to compare by, or a custom prompt template. The
        comparison task can be framed in a number of ways (choosing the greater item, lesser item, or the ordering).

        Args:
            a: The first item to compare.
            b: The second item to compare.
            by: A criteria specifying what aspect to compare by. If this is provided, `template` cannot be
                provided.
            to_str: If specified, used to convert items to string representation. Otherewise, uses `str()` on each item.
                If this is provided, a callable template cannot be provided.
            template: A custom prompt template for the comparison. Must be either a string template with two positional
                placeholders, or a callable that takes two items and returns a formatted string. If this is provided,
                `by` cannot be provided.
            task: The type of comparison task that is being performed in `template`. This allows for writing the
                template in the most convenient way possible (e.g., in some scenarios, it's easier to specify a criteria
                for which item is lesser, and in others, it's easier to specify a criteria for which item is greater).
                If this is provided, a custom `template` must also be provided.  Defaults to
                [Task.CHOOSE_GREATER][semlib.compare.Task.CHOOSE_GREATER] if not specified.
            model: If specified, overrides the default model for this call.

        Returns:
            The ordering of the two items.

        Raises:
            ValidationError: If parsing the LLM response fails.

        Examples:
            Basic comparison:
            >>> await session.compare("twelve", "seventy two")
            <Order.LESS: 'less'>

            Custom criteria:
            >>> await session.compare("California condor", "Bald eagle", by="wingspan")
            <Order.GREATER: 'greater'>

            Custom template and task:
            >>> await session.compare(
            ...     "proton",
            ...     "electron",
            ...     template="Which is smaller, (A) {} or (B) {}?",
            ...     task=Task.CHOOSE_LESSER,
            ... )
            <Order.GREATER: 'greater'>
        """
        if task is not None and task not in {Task.CHOOSE_GREATER, Task.CHOOSE_GREATER_OR_ABSTAIN} and template is None:
            msg = "if 'task' is not CHOOSE_GREATER or CHOOSE_GREATER_OR_ABSTAIN, 'template' must also be provided"
            raise ValueError(msg)
        if template is not None:
            if callable(template) and to_str is not None:
                msg = "cannot provide 'to_str' when a template function is provided"
                raise ValueError(msg)
            if by is not None:
                msg = "cannot provide 'by' when a custom template is provided"
                raise ValueError(msg)

        to_str = to_str if to_str is not None else str
        if task is None:
            task = Task.CHOOSE_GREATER
        elif isinstance(task, str):
            task = Task(task)
        model = model if model is not None else self._model

        if isinstance(template, str):
            prompt = template.format(to_str(a), to_str(b))
        elif template is not None:
            # callable
            prompt = template(a, b)
        elif by is None:
            prompt = _DEFAULT_TEMPLATE.format(a=to_str(a), b=to_str(b))
        else:
            prompt = _DEFAULT_TEMPLATE_BY.format(criteria=by, a=to_str(a), b=to_str(b))

        response = await self.prompt(
            prompt,
            model=model,
            return_type=_RETURN_TYPE_BY_TASK[task],
        )

        match task:
            case Task.COMPARE:
                strict_compare_result = cast(_StrictCompareResult, response)
                return strict_compare_result.order.to_order()
            case Task.COMPARE_OR_ABSTAIN:
                compare_result = cast(_CompareResult, response)
                return compare_result.order
            case Task.CHOOSE_GREATER:
                strict_choose_result = cast(_StrictChooseResult, response)
                match strict_choose_result.choice:
                    case _StrictChoice.A:
                        return Order.GREATER
                    case _StrictChoice.B:
                        return Order.LESS
            case Task.CHOOSE_GREATER_OR_ABSTAIN:
                choose_result = cast(_ChooseResult, response)
                match choose_result.choice:
                    case _Choice.A:
                        return Order.GREATER
                    case _Choice.B:
                        return Order.LESS
                    case _Choice.NEITHER:
                        return Order.NEITHER
            case Task.CHOOSE_LESSER:
                strict_choose_result = cast(_StrictChooseResult, response)
                match strict_choose_result.choice:
                    case _StrictChoice.A:
                        return Order.LESS
                    case _StrictChoice.B:
                        return Order.GREATER
            case Task.CHOOSE_LESSER_OR_ABSTAIN:
                choose_result = cast(_ChooseResult, response)
                match choose_result.choice:
                    case _Choice.A:
                        return Order.LESS
                    case _Choice.B:
                        return Order.GREATER
                    case _Choice.NEITHER:
                        return Order.NEITHER


async def compare[T](
    a: T,
    b: T,
    /,
    *,
    by: str | None = None,
    to_str: Callable[[T], str] | None = None,
    template: str | Callable[[T, T], str] | None = None,
    task: Task | str | None = None,
    model: str | None = None,
) -> Order:
    """Standalone version of [compare][semlib.compare.Compare.compare]."""
    comparator = Compare(model=model)
    return await comparator.compare(a, b, by=by, to_str=to_str, template=template, task=task)


def compare_sync[T](
    a: T,
    b: T,
    /,
    *,
    by: str | None = None,
    to_str: Callable[[T], str] | None = None,
    template: str | Callable[[T, T], str] | None = None,
    task: Task | str | None = None,
    model: str | None = None,
) -> Order:
    """Standalone synchronous version of [compare][semlib.compare.Compare.compare]."""
    comparator = Compare(model=model)
    return asyncio.run(comparator.compare(a, b, by=by, to_str=to_str, template=template, task=task))



from collections.abc import Callable, Coroutine, Iterable
from typing import override

from semlib._internal import util
from semlib._internal.util import foreach
from semlib.compare import Order
from semlib.sort.algorithm.algorithm import Algorithm


class BordaCount(Algorithm):
    """Borda count sorting algorithm.

    This algorithm uses the [Borda count](https://en.wikipedia.org/wiki/Borda_count) method to rank items.  The
    algorithm has good theoretical properties for finding approximate rankings based on noisy pairwise comparisons
    ([Shah and Wainwright, 2018](https://jmlr.org/papers/volume18/16-206/16-206.pdf)).

    This algorithm requires O(n^2) pairwise comparisons. If you want to reduce the number of comparisons (to reduce LLM
    costs), you can consider using the [QuickSort][semlib.sort.algorithm.QuickSort] algorithm instead.

    This algorithm is carefully implemented so that it has O(n) space complexity.
    """

    @override
    async def _sort[T](
        self,
        iterable: Iterable[T],
        /,
        *,
        reverse: bool = False,
        comparator: Callable[[T, T], Coroutine[None, None, Order]],
        max_concurrency: int,
    ) -> list[T]:
        lst = iterable if isinstance(iterable, list) else list(iterable)
        scores: list[int] = [0 for _ in lst]

        async def fn(item: tuple[int, int]) -> None:
            i, j = item
            result_ij, result_ji = await util.gather(comparator(lst[i], lst[j]), comparator(lst[j], lst[i]))
            if result_ij == Order.LESS and result_ji == Order.GREATER:
                scores[j] += 1
                scores[i] -= 1
            elif result_ij == Order.GREATER and result_ji == Order.LESS:
                scores[i] += 1
                scores[j] -= 1

        await foreach(
            fn,
            ((i, j) for i in range(len(lst)) for j in range(i + 1, len(lst))),
            max_concurrency=max(1, max_concurrency // 2),  # because each worker does two comparisons concurrently
        )

        # stable sort
        sort_by_score = sorted([(scores[i], i, lst[i]) for i in range(len(lst))], reverse=reverse)
        return [item for _, _, item in sort_by_score]



import pytest

from semlib._internal.util import foreach


@pytest.mark.asyncio
async def test_foreach_raise() -> None:
    async def fn(_: int) -> None:
        msg = "foo"
        raise ValueError(msg)

    with pytest.raises(ValueError, match="foo"):
        await foreach(fn, range(10), max_concurrency=2)



from collections.abc import Callable

import pydantic
import pytest

from semlib.bare import Bare
from semlib.box import Box
from semlib.reduce import reduce, reduce_sync
from tests.conftest import LLMMocker


def test_reduce_str(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "What is 0 + 1? Respond with just the number.": "1",
            "What is 1 + 2? Respond with just the number.": "3",
            "What is 3 + 3? Respond with just the number.": "6",
            "What is 6 + 4? Respond with just the number.": "10",
        }
    )
    with mocker.patch_llm():
        result: str = reduce_sync(map(str, range(5)), template="What is {} + {}? Respond with just the number.")
    assert result == "10"


def test_reduce_str_associative(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "What is 0 + 1? Respond with just the number.": "1",
            "What is 3 + 4? Respond with just the number.": "7",
            "What is 2 + 7? Respond with just the number.": "9",
            "What is 1 + 9? Respond with just the number.": "10",
        }
    )
    with mocker.patch_llm():
        result = reduce_sync(
            map(str, range(5)), template="What is {} + {}? Respond with just the number.", associative=True
        )
    assert result == "10"


@pytest.mark.asyncio
@pytest.mark.parametrize("associative", [True, False])
async def test_reduce_str_fn(llm_mocker: Callable[[dict[str, str]], LLMMocker], associative: bool) -> None:  # noqa: FBT001
    mocker = llm_mocker(
        {
            "What is 0 + 1? Respond with just the number.": "1",
            "What is 1 + 2? Respond with just the number.": "3",
            "What is 3 + 3? Respond with just the number.": "6",
            "What is 6 + 4? Respond with just the number.": "10",
            "What is 3 + 4? Respond with just the number.": "7",
            "What is 2 + 7? Respond with just the number.": "9",
            "What is 1 + 9? Respond with just the number.": "10",
        }
    )
    with mocker.patch_llm():
        result = await reduce(
            map(str, range(5)),
            template=lambda a, b: f"What is {a} + {b}? Respond with just the number.",
            associative=associative,
        )
    assert result == "10"


@pytest.mark.asyncio
async def test_reduce_t_fn(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "What is 0 + 1? Respond with just the number.": '{"value": 1}',
            "What is 1 + 2? Respond with just the number.": '{"value": 3}',
            "What is 3 + 3? Respond with just the number.": '{"value": 6}',
            "What is 6 + 4? Respond with just the number.": '{"value": 10}',
        }
    )

    class Number(pydantic.BaseModel):
        value: int

    with mocker.patch_llm():
        result: Number = await reduce(
            (Number(value=n) for n in range(5)),
            template=lambda a, b: f"What is {a.value} + {b.value}? Respond with just the number.",
            return_type=Number,
            associative=False,
        )

    assert result == Number(value=10)


@pytest.mark.asyncio
async def test_reduce_t_fn_associative(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "What is 0 + 1? Respond with just the number.": '{"value": 1}',
            "What is 3 + 4? Respond with just the number.": '{"value": 7}',
            "What is 2 + 7? Respond with just the number.": '{"value": 9}',
            "What is 1 + 9? Respond with just the number.": '{"value": 10}',
        }
    )

    class Number(pydantic.BaseModel):
        value: int

    with mocker.patch_llm():
        result: Number = await reduce(
            (Number(value=n) for n in range(5)),
            template=lambda a, b: f"What is {a.value} + {b.value}? Respond with just the number.",
            return_type=Number,
            associative=True,
        )

    assert result == Number(value=10)


@pytest.mark.asyncio
@pytest.mark.parametrize("associative", [True, False])
async def test_reduce_t(llm_mocker: Callable[[dict[str, str]], LLMMocker], associative: bool) -> None:  # noqa: FBT001
    mocker = llm_mocker(
        {
            "What is value=0 + value=1? Respond with just the number.": '{"value": 1}',
            "What is value=1 + value=2? Respond with just the number.": '{"value": 3}',
            "What is value=3 + value=3? Respond with just the number.": '{"value": 6}',
            "What is value=6 + value=4? Respond with just the number.": '{"value": 10}',
            "What is value=3 + value=4? Respond with just the number.": '{"value": 7}',
            "What is value=2 + value=7? Respond with just the number.": '{"value": 9}',
            "What is value=1 + value=9? Respond with just the number.": '{"value": 10}',
        }
    )

    class Number(pydantic.BaseModel):
        value: int

    with mocker.patch_llm():
        result: Number = await reduce(
            (Number(value=n) for n in range(5)),
            template="What is {} + {}? Respond with just the number.",
            return_type=Number,
            associative=associative,
        )

    assert result == Number(value=10)


def test_reduce_tu(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "If 0 is prime, append it to this list: [].": '{"items": []}',
            "If 1 is prime, append it to this list: [].": '{"items": []}',
            "If 2 is prime, append it to this list: [].": '{"items": [2]}',
            "If 3 is prime, append it to this list: [2].": '{"items": [2, 3]}',
            "If 4 is prime, append it to this list: [2, 3].": '{"items": [2, 3]}',
            "If 5 is prime, append it to this list: [2, 3].": '{"items": [2, 3, 5]}',
            "If 6 is prime, append it to this list: [2, 3, 5].": '{"items": [2, 3, 5]}',
            "If 7 is prime, append it to this list: [2, 3, 5].": '{"items": [2, 3, 5, 7]}',
            "If 8 is prime, append it to this list: [2, 3, 5, 7].": '{"items": [2, 3, 5, 7]}',
            "If 9 is prime, append it to this list: [2, 3, 5, 7].": '{"items": [2, 3, 5, 7]}',
        }
    )

    class List(pydantic.BaseModel):
        items: list[int]

    with mocker.patch_llm():
        result = reduce_sync(
            range(10),
            template=lambda acc, n: f"If {n} is prime, append it to this list: {acc.items}.",
            initial=List(items=[]),
            return_type=List,
        )

    assert result == List(items=[2, 3, 5, 7])


@pytest.mark.asyncio
async def test_reduce_tu2(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "Add Stanford to this set of universities: [].": '{"universities": ["Stanford"]}',
            "Add MIT to this set of universities: ['Stanford'].": '{"universities": ["Stanford", "MIT"]}',
            "Add Harvard to this set of universities: ['Stanford', 'MIT'].": '{"universities": ["Stanford", "MIT", "Harvard"]}',
            "Add MIT to this set of universities: ['Stanford', 'MIT', 'Harvard'].": '{"universities": ["Stanford", "MIT", "Harvard"]}',
        }
    )

    class Person(pydantic.BaseModel):
        name: str
        university: str

    people = [
        Person(name="David Mazieres", university="Stanford"),
        Person(name="Frans Kaashoek", university="MIT"),
        Person(name="Eddie Kohler", university="Harvard"),
        Person(name="Nickolai Zeldovich", university="MIT"),
    ]

    class Universities(pydantic.BaseModel):
        universities: list[str]

    with mocker.patch_llm():
        result = await reduce(
            people,
            template=lambda acc, p: f"Add {p.university} to this set of universities: {acc.universities}.",
            initial=Universities(universities=[]),
            return_type=Universities,
        )

    assert result == Universities(universities=["Stanford", "MIT", "Harvard"])


def test_reduce_bare(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "What is 1 + 2? Respond with just the number.": '{"value": 3}',
            "What is 3 + 4? Respond with just the number.": '{"value": 7}',
            "What is 3 + 7? Respond with just the number.": '{"value": 10}',
        }
    )

    def template(a: int, b: int) -> str:
        return f"What is {a} + {b}? Respond with just the number."

    with mocker.patch_llm():
        result = reduce_sync(
            [1, 2, 3, 4],
            template=template,
            associative=True,
            return_type=Bare(int),
        )
    assert result == 10


def test_reduce_bare_2(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "If 1 is prime, append it to this list: [].": '{"value": []}',
            "If 2 is prime, append it to this list: [].": '{"value": [2]}',
            "If 3 is prime, append it to this list: [2].": '{"value": [2, 3]}',
            "If 4 is prime, append it to this list: [2, 3].": '{"value": [2, 3]}',
            "If 5 is prime, append it to this list: [2, 3].": '{"value": [2, 3, 5]}',
            "If 6 is prime, append it to this list: [2, 3, 5].": '{"value": [2, 3, 5]}',
        }
    )

    def template(acc: list[int], n: int) -> str:
        return f"If {n} is prime, append it to this list: {acc}."

    with mocker.patch_llm():
        result = reduce_sync(
            [1, 2, 3, 4, 5, 6],
            template=template,
            return_type=Bare(list[int]),
            initial=[],
        )
    assert result == [2, 3, 5]


def test_reduce_raises() -> None:
    with pytest.raises(ValueError, match="empty iterable"):
        reduce_sync([], template="What is {} + {}?")
    with pytest.raises(ValueError, match="empty iterable"):
        reduce_sync([], template="What is {} + {}?", associative=True)

    class Number(pydantic.BaseModel):
        value: int

    with pytest.raises(ValueError, match="empty iterable"):
        reduce_sync([], template="What is {} + {}?", return_type=Number)
    with pytest.raises(ValueError, match="empty iterable"):
        reduce_sync([], template="What is {} + {}?", return_type=Number, associative=True)


def test_reduce_box(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "Construct a set from the numbers '1' and 'two'.": "{1, 2}",
            "Construct a set from the numbers '2' and 'three'.": "{2, 3}",
            "Add the number '3' to the set {2, 3}.": "{2, 3}",
            "Compute the union of the sets {1, 2} and {2, 3}.": "{1, 2, 3}",
        }
    )

    items: list[str] = ["1", "two", "3", "2", "three"]

    def template(a: Box[str] | str, b: Box[str] | str) -> str:
        if isinstance(a, Box) and isinstance(b, Box):
            return f"Construct a set from the numbers '{a.value}' and '{b.value}'."
        if not isinstance(a, Box) and not isinstance(b, Box):
            return f"Compute the union of the sets {a} and {b}."
        if isinstance(a, Box):
            a, b = b, a
        assert isinstance(b, Box)
        assert isinstance(a, str)
        return f"Add the number '{b.value}' to the set {a}."

    with mocker.patch_llm():
        result: Box[str] | str = reduce_sync(map(Box, items), template=template, associative=True)
    assert result == "{1, 2, 3}"



import asyncio
import datetime
from collections.abc import Callable
from typing import Any
from unittest.mock import AsyncMock, patch

import pydantic
import pytest
from litellm.types.utils import Message

from semlib._internal.base import Base
from semlib._internal.util import gather
from semlib.bare import Bare
from semlib.cache import InMemoryCache
from tests.conftest import LLMMocker


@pytest.mark.asyncio
async def test_base_no_cache(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    base = Base()
    assert base.total_cost() == 0.0

    mocker = llm_mocker(
        {
            "What is 1 + 1?": "2",
            "What is 2 + 2?": "4",
        }
    )

    with mocker.patch_llm():
        result1: str = await base._acompletion(messages=[Message(role="user", content="What is 1 + 1?")])  # noqa: SLF001
        assert result1 == "2"
        assert base.total_cost() == 1.0
        result2: str = await base._acompletion(messages=[Message(role="user", content="What is 2 + 2?")])  # noqa: SLF001
        assert result2 == "4"
        assert base.total_cost() == 2.0
        # repeat query, still incurs cost because no cache
        result2 = await base._acompletion(messages=[Message(role="user", content="What is 2 + 2?")])  # noqa: SLF001
        assert result2 == "4"
        assert base.total_cost() == 3.0


@pytest.mark.asyncio
async def test_base_cache(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    base = Base(cache=InMemoryCache())
    assert base.total_cost() == 0.0

    mocker = llm_mocker(
        {
            "What is 1 + 1?": "2",
            "What is 2 + 2?": "4",
        }
    )

    with mocker.patch_llm():
        result1: str = await base._acompletion(messages=[Message(role="user", content="What is 1 + 1?")])  # noqa: SLF001
        assert result1 == "2"
        assert base.total_cost() == 1.0
        result2: str = await base._acompletion(messages=[Message(role="user", content="What is 2 + 2?")])  # noqa: SLF001
        assert result2 == "4"
        assert base.total_cost() == 2.0
        # repeat query, doesn't incur additional cost
        result2 = await base._acompletion(messages=[Message(role="user", content="What is 2 + 2?")])  # noqa: SLF001
        assert result2 == "4"
        assert base.total_cost() == 2.0
        # but it does when the model is different
        result2 = await base._acompletion(  # noqa: SLF001
            messages=[Message(role="user", content="What is 2 + 2?")], model="gpt-4.1-nano"
        )
        assert result2 == "4"
        assert base.total_cost() == 3.0


@pytest.mark.asyncio
async def test_base_return_type(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    class Person(pydantic.BaseModel):
        name: str
        date_of_birth: datetime.date

    base = Base()
    assert base.total_cost() == 0.0

    mocker = llm_mocker(
        {
            "Who was the 40th president of the United States?": '{"name": "Ronald Reagan", "date_of_birth": "1911-02-06"}',
            "Who was the 41th president of the United States?": '{"name": "George H. W. Bush", "date_of_birth": "1924-06-12"}',
        }
    )

    with mocker.patch_llm():
        result1: Person = await base._acompletion(  # noqa: SLF001
            messages=[Message(role="user", content="Who was the 40th president of the United States?")],
            return_type=Person,
        )
        assert result1 == Person(name="Ronald Reagan", date_of_birth=datetime.date(1911, 2, 6))
        assert base.total_cost() == 1.0
        result2: Person = await base._acompletion(  # noqa: SLF001
            messages=[Message(role="user", content="Who was the 41th president of the United States?")],
            return_type=Person,
        )
        assert result2 == Person(name="George H. W. Bush", date_of_birth=datetime.date(1924, 6, 12))
        assert base.total_cost() == 2.0


@pytest.mark.asyncio
async def test_base_prompt(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "What is 1 + 1?": "2",
        }
    )
    base = Base()
    with mocker.patch_llm():
        response = await base.prompt("What is 1 + 1?")
    assert response == "2"


@pytest.mark.asyncio
async def test_base_prompt_typed(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    class Number(pydantic.BaseModel):
        value: int

    mocker = llm_mocker(
        {
            "What is 1 + 1?": '{"value": 2}',
        }
    )
    base = Base()
    with mocker.patch_llm():
        response = await base.prompt("What is 1 + 1?", return_type=Number)
    assert response == Number(value=2)


def test_base_ollama() -> None:
    base = Base(model="ollama/gemma3:12b")
    assert base._max_concurrency == 1  # noqa: SLF001


def test_base_bad_args() -> None:
    with pytest.raises(ValueError, match="must be a positive integer"):
        Base(max_concurrency=-3)


@pytest.mark.asyncio
async def test_base_raises(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    base = Base()

    mocker = llm_mocker(
        {
            "What is 1 + 1?": "2",
        }
    )

    class Number(pydantic.BaseModel):
        value: int

    with mocker.patch_llm(), pytest.raises(pydantic.ValidationError):
        await base._acompletion(  # noqa: SLF001
            messages=[Message(role="user", content="What is 1 + 1?")],
            return_type=Number,
        )


@pytest.mark.asyncio
async def test_prompt_bare_cached(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "Populate the list": '{"primes": [2, 3, 5]}',
        }
    )
    b: Bare[list[int]] = Bare(list[int], class_name="list_of_first_three", field_name="primes")
    base = Base(cache=InMemoryCache())
    with mocker.patch_llm():
        x: list[int] = await base.prompt("Populate the list", return_type=b)
        x = await base.prompt("Populate the list", return_type=b)
    assert x == [2, 3, 5]
    assert base.total_cost() == 1.0


def test_base_env_max_concurrency() -> None:
    import os

    os.environ["SEMLIB_MAX_CONCURRENCY"] = "7"
    base = Base()
    assert base._max_concurrency == 7  # noqa: SLF001

    os.environ["SEMLIB_MAX_CONCURRENCY"] = "not_an_int"
    with pytest.raises(ValueError, match="must be an integer"):
        Base()

    del os.environ["SEMLIB_MAX_CONCURRENCY"]


@pytest.mark.asyncio
async def test_pending_requests_deduplication() -> None:
    """Test that concurrent identical requests wait for the first to complete instead of making duplicate API calls.

    This test relies on timing / asyncio.sleep to catch buggy implementations, but with a correct implementation, it
    should pass regardless of thread scheduling / timing. I don't think there's a good way to test this
    deterministically without glass-box testing. In practice, this should effectively catch buggy implementations (e.g.,
    the Semlib code pre-deduplication).
    """
    call_count = 0

    async def mock_acompletion(*_args: Any, **_kwargs: Any) -> Any:
        nonlocal call_count
        call_count += 1
        await asyncio.sleep(0.2)  # simulate slow completion
        response = AsyncMock()
        response.choices = [AsyncMock()]
        response.choices[0].message.content = "42"
        return response

    base = Base(cache=InMemoryCache())

    async def coro() -> str:
        return await base._acompletion(messages=[Message(role="user", content="What is the answer?")])  # noqa: SLF001

    with patch("litellm.acompletion", side_effect=mock_acompletion), patch("litellm.completion_cost", return_value=1.0):
        res1, res2 = await gather(coro(), coro())
        assert res1 == "42"
        assert res2 == "42"
        assert call_count == 1
        assert base.total_cost() == 1.0



from collections.abc import Callable

import pydantic
import pytest

from semlib import InMemoryCache, Session
from semlib._internal.constants import DEFAULT_MODEL
from semlib.compare import _DEFAULT_TEMPLATE
from tests.conftest import LLMMocker


@pytest.mark.asyncio
async def test_session(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    session = Session(cache=InMemoryCache())

    measurements: list[str] = ["1 mile", "1.4 kilometers", "1 foot", "3 inches", "3 centimeters"]

    mocker = llm_mocker(
        {
            _DEFAULT_TEMPLATE.format(a="1 mile", b="1.4 kilometers"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a="1 mile", b="1 foot"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a="1 mile", b="3 inches"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a="1 mile", b="3 centimeters"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a="1.4 kilometers", b="1 foot"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a="1.4 kilometers", b="3 inches"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a="1.4 kilometers", b="3 centimeters"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a="1 foot", b="3 inches"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a="1 foot", b="3 centimeters"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(a="3 inches", b="3 centimeters"): '{"choice": "A"}',
            _DEFAULT_TEMPLATE.format(b="1 mile", a="1.4 kilometers"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(b="1 mile", a="1 foot"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(b="1 mile", a="3 inches"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(b="1 mile", a="3 centimeters"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(b="1.4 kilometers", a="1 foot"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(b="1.4 kilometers", a="3 inches"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(b="1.4 kilometers", a="3 centimeters"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(b="1 foot", a="3 inches"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(b="1 foot", a="3 centimeters"): '{"choice": "B"}',
            _DEFAULT_TEMPLATE.format(b="3 inches", a="3 centimeters"): '{"choice": "B"}',
            "Convert the length 1 mile to centimeters, rounding to the nearest integer.": '{"centimeters": 160934}',
            "Convert the length 1.4 kilometers to centimeters, rounding to the nearest integer.": '{"centimeters": 140000}',
            "Convert the length 1 foot to centimeters, rounding to the nearest integer.": '{"centimeters": 30}',
            "Convert the length 3 inches to centimeters, rounding to the nearest integer.": '{"centimeters": 8}',
            "Convert the length 3 centimeters to centimeters, rounding to the nearest integer.": '{"centimeters": 3}',
        }
    )

    with mocker.patch_llm():
        sort_measurements: list[str] = await session.sort(measurements)

    assert sort_measurements == ["3 centimeters", "3 inches", "1 foot", "1.4 kilometers", "1 mile"]
    assert session.total_cost() == 20

    # running again should hit the cache and incur no additional cost
    with mocker.patch_llm():
        sort_measurements = await session.sort(measurements)
    assert session.total_cost() == 20

    # unless we clear the cache
    session.clear_cache()
    with mocker.patch_llm():
        sort_measurements = await session.sort(measurements)
    assert session.total_cost() == 40

    class Length(pydantic.BaseModel):
        centimeters: int

    with mocker.patch_llm():
        measurements_cm: list[Length] = await session.map(
            sort_measurements,
            template="Convert the length {} to centimeters, rounding to the nearest integer.",
            return_type=Length,
        )
    assert session.total_cost() == 45

    assert measurements_cm == [
        Length(centimeters=3),
        Length(centimeters=8),
        Length(centimeters=30),
        Length(centimeters=140_000),
        Length(centimeters=160_934),
    ]

    assert session.model == DEFAULT_MODEL



import asyncio
from collections.abc import Callable
from typing import Any
from unittest.mock import AsyncMock, patch

import pydantic
import pytest

from semlib.filter import _DEFAULT_TEMPLATE
from semlib.find import Find, find, find_sync
from tests.conftest import LLMMocker


@pytest.mark.parametrize("negate", [True, False])
def test_find(llm_mocker: Callable[[dict[str, str]], LLMMocker], negate: bool) -> None:  # noqa: FBT001
    people = ["Barack Obama", "Miley Cyrus", "Jeff Dean", "Jeff Bezos", "Bill Clinton", "Elon Musk"]
    mocker = llm_mocker(
        {
            _DEFAULT_TEMPLATE.format(item="Barack Obama", by="singer?"): '{"decision": false}',
            _DEFAULT_TEMPLATE.format(item="Miley Cyrus", by="singer?"): '{"decision": true}',
            _DEFAULT_TEMPLATE.format(item="Jeff Dean", by="singer?"): '{"decision": false}',
            _DEFAULT_TEMPLATE.format(item="Jeff Bezos", by="singer?"): '{"decision": false}',
            _DEFAULT_TEMPLATE.format(item="Bill Clinton", by="singer?"): '{"decision": false}',
            _DEFAULT_TEMPLATE.format(item="Elon Musk", by="singer?"): '{"decision": false}',
        }
    )
    with mocker.patch_llm():
        result = find_sync(people, by="singer?", negate=negate)

    if negate:
        assert result != "Miley Cyrus"
    else:
        assert result == "Miley Cyrus"


def test_find_not_found(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    people = ["Miley Cyrus", "Jeff Dean"]
    mocker = llm_mocker(
        {
            _DEFAULT_TEMPLATE.format(item="Miley Cyrus", by="president of the United States"): '{"decision": false}',
            _DEFAULT_TEMPLATE.format(item="Jeff Dean", by="president of the United States"): '{"decision": false}',
        }
    )
    with mocker.patch_llm():
        result = find_sync(people, by="president of the United States")
    assert result is None


def test_find_template_str(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    people = ["Barack Obama", "Miley Cyrus", "Jeff Dean", "Jeff Bezos", "Bill Clinton", "Elon Musk"]
    mocker = llm_mocker(
        {
            "Is Barack Obama a famous computer scientist?": '{"decision": false}',
            "Is Miley Cyrus a famous computer scientist?": '{"decision": false}',
            "Is Jeff Dean a famous computer scientist?": '{"decision": true}',
            "Is Jeff Bezos a famous computer scientist?": '{"decision": false}',
            "Is Bill Clinton a famous computer scientist?": '{"decision": false}',
            "Is Elon Musk a famous computer scientist?": '{"decision": false}',
        }
    )
    with mocker.patch_llm():
        result = find_sync(people, template="Is {} a famous computer scientist?")
    assert result == "Jeff Dean"


@pytest.mark.asyncio
async def test_find_callable(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    pairs = [
        (123, 321),
        (1838, 1883),
        (33, 44),
        (12, 21),
    ]
    mocker = llm_mocker(
        {
            "Is 123 the reverse of 321?": '{"decision": true}',
            "Is 12 the reverse of 21?": '{"decision": true}',
            "Is 1838 the reverse of 1883?": '{"decision": false}',
            "Is 33 the reverse of 44?": '{"decision": false}',
        }
    )
    with mocker.patch_llm():
        result = await find(pairs, template=lambda pair: f"Is {pair[0]} the reverse of {pair[1]}?")
    assert result in {(123, 321), (12, 21)}


@pytest.mark.asyncio
async def test_find_cancellation() -> None:
    async def deadlock() -> None:
        q: asyncio.Queue[None] = asyncio.Queue()
        await q.get()

    async def mock_acompletion(*_args: Any, **kwargs: Any) -> Any:
        if kwargs.get("messages", [])[0].get("content", "") == "Is 1 odd?":
            response = AsyncMock()
            response.choices = [AsyncMock()]
            response.choices[0].message.content = '{"decision": true}'
            return response
        await deadlock()
        return None

    finder = Find(max_concurrency=2)
    with patch("litellm.acompletion", side_effect=mock_acompletion), patch("litellm.completion_cost", return_value=1.0):
        result = await finder.find([1, 2, 3, 4, 5], template="Is {} odd?")
    assert result == 1
    assert finder.total_cost() == 1.0


def test_find_bad_args() -> None:
    with pytest.raises(ValueError, match="must specify either"):
        find_sync([1, 2, 3])
    with pytest.raises(ValueError, match="cannot provide"):
        find_sync([1, 2, 3], template=lambda x: f"Is {x} odd?", to_str=str)
    with pytest.raises(ValueError, match="cannot provide"):
        find_sync([1, 2, 3], by="odd", template="Is {} odd?")


def test_find_raises(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    people = ["Barack Obama", "Miley Cyrus", "Jeff Dean"]
    mocker = llm_mocker(
        {
            _DEFAULT_TEMPLATE.format(item="Barack Obama", by="president of the United States"): "not json",
            _DEFAULT_TEMPLATE.format(item="Miley Cyrus", by="president of the United States"): '{"decision": false}',
            _DEFAULT_TEMPLATE.format(item="Jeff Dean", by="president of the United States"): '{"decision": false}',
        }
    )

    with mocker.patch_llm(), pytest.raises(pydantic.ValidationError):
        find_sync(people, by="president of the United States")


@pytest.mark.asyncio
async def test_find_task_cleanup_on_exception() -> None:
    task_cancelled = asyncio.Event()

    async def mock_acompletion(*_args: Any, **kwargs: Any) -> Any:
        content = kwargs.get("messages", [{}])[0].get("content", "")

        if "item_1" in content:
            # first task fails after a short delay
            await asyncio.sleep(0.1)
            msg = "Mock API error"
            raise ValueError(msg)

        # "item_2" case: other task takes a long time to complete
        try:
            q: asyncio.Queue[None] = asyncio.Queue()
            await q.get()
        except asyncio.CancelledError:
            task_cancelled.set()
            raise

    with patch("litellm.acompletion", side_effect=mock_acompletion), patch("litellm.completion_cost", return_value=0.0):
        with pytest.raises(ValueError, match="Mock API error"):
            await find(["item_1", "item_2"], by="test")

        await task_cancelled.wait()  # this will hang if the task wasn't cancelled



import asyncio
from typing import overload

from pydantic import BaseModel

from semlib._internal.base import Base
from semlib.bare import Bare


@overload
async def prompt[T: BaseModel](prompt: str, /, *, return_type: type[T], model: str | None = None) -> T: ...


@overload
async def prompt[T](prompt: str, /, *, return_type: Bare[T], model: str | None = None) -> T: ...


@overload
async def prompt(prompt: str, /, *, return_type: None = None, model: str | None = None) -> str: ...


async def prompt[T: BaseModel, U](
    prompt: str, /, *, return_type: type[T] | Bare[U] | None = None, model: str | None = None
) -> str | T | U:
    """Standalone version of [prompt][semlib._internal.base.Base.prompt]."""
    base = Base(model=model)
    return await base.prompt(prompt, return_type=return_type)


@overload
def prompt_sync[T: BaseModel](prompt: str, /, *, return_type: type[T], model: str | None = None) -> T: ...


@overload
def prompt_sync[T](prompt: str, /, *, return_type: Bare[T], model: str | None = None) -> T: ...


@overload
def prompt_sync(prompt: str, /, *, return_type: None = None, model: str | None = None) -> str: ...


def prompt_sync[T: BaseModel, U](
    prompt: str, /, *, return_type: type[T] | Bare[U] | None = None, model: str | None = None
) -> str | T | U:
    """Standalone synchronous version of [prompt][semlib._internal.base.Base.prompt]."""
    base = Base(model=model)
    return asyncio.run(base.prompt(prompt, return_type=return_type))  # type: ignore[return-value]



import pytest

from semlib.bare import Bare


def test_bare_invalid_extract() -> None:
    b = Bare(int)
    with pytest.raises(TypeError, match="expected instance of"):
        b._extract("hi")  # noqa: SLF001



import asyncio
from collections.abc import Callable, Iterable
from typing import Any, overload

from pydantic import BaseModel
from pydantic_core import ValidationError  # noqa: F401 # used in docstrings

from semlib._internal.base import Base
from semlib._internal.util import gather
from semlib.bare import Bare


class Reduce(Base):
    @overload
    async def reduce(
        self,
        iterable: Iterable[str],
        /,
        template: str | Callable[[str, str], str],
        *,
        associative: bool = False,
        model: str | None = None,
    ) -> str: ...

    @overload
    async def reduce[T](
        self,
        iterable: Iterable[str | T],
        /,
        template: str | Callable[[str | T, str | T], str],
        *,
        associative: bool = False,
        model: str | None = None,
    ) -> str | T: ...

    @overload
    async def reduce[T: BaseModel](
        self,
        iterable: Iterable[T],
        /,
        template: str | Callable[[T, T], str],
        *,
        return_type: type[T],
        associative: bool = False,
        model: str | None = None,
    ) -> T: ...

    @overload
    async def reduce[T](
        self,
        iterable: Iterable[T],
        /,
        template: str | Callable[[T, T], str],
        *,
        return_type: Bare[T],
        associative: bool = False,
        model: str | None = None,
    ) -> T: ...

    @overload
    async def reduce[T, U: BaseModel](
        self,
        iterable: Iterable[T],
        /,
        template: str | Callable[[U, T], str],
        initial: U,
        *,
        return_type: type[U],
        model: str | None = None,
    ) -> U: ...

    @overload
    async def reduce[T, U](
        self,
        iterable: Iterable[T],
        /,
        template: str | Callable[[U, T], str],
        initial: U,
        *,
        return_type: Bare[U],
        model: str | None = None,
    ) -> U: ...

    async def reduce(
        self,
        iterable: Iterable[Any],
        /,
        template: str | Callable[[Any, Any], str],
        initial: Any = None,
        *,
        return_type: Any = None,
        associative: bool = False,
        model: str | None = None,
    ) -> Any:
        """Reduce an iterable to a single value using a language model.

        This method is analogous to Python's
        [`functools.reduce`](https://docs.python.org/3/library/functools.html#functools.reduce) function.

        Args:
            iterable: The collection of items to reduce.
            template: A prompt template to apply to each item. This can be either a string template with two positional
                placeholders (with the first placeholder being the accumulator and the second placeholder being an
                item), or a callable that takes an accumulator and an item and returns a formatted string.
            initial: If provided, this value is placed before the items of the iterable in the calculation, and serves
                as a default when the iterable is empty.
            return_type: The return type is also the type of the accumulator. If not specified, the responses are
                returned as raw strings. If a Pydantic model class is provided, the responses are parsed into instances
                of that model. If a [Bare][semlib.bare.Bare] instance is provided, single values of the specified type
                are extracted from the responses.
            associative: If `True`, the reduction is performed in a balanced tree manner, which unlocks concurrency and
                can provide significant speedups for large iterables. This requires the reduction operation to be
                associative.
            model: If specified, overrides the default model for this call.

        Returns:
            The final accumulated value.

        Raises:
            ValidationError: If `return_type` is a Pydantic model or a Bare type and the response cannot be
                parsed into the specified type.

        Examples:
            Basic reduce:
            >>> await session.reduce(
            ...     ["one", "three", "seven", "twelve"], "{} + {} = ?", return_type=Bare(int)
            ... )
            23

            Reduce with initial value:
            >>> await session.reduce(
            ...     range(20),
            ...     template=lambda acc, n: f"If {n} is prime, append it to this list: {acc}.",
            ...     initial=[],
            ...     return_type=Bare(list[int]),
            ...     model="openai/o4-mini",
            ... )
            [2, 3, 5, 7, 11, 13, 17, 19]

            Associative reduce:
            >>> await session.reduce(
            ...     [[i] for i in range(20)],
            ...     template=lambda acc,
            ...     n: f"Compute the union of these two sets, and then remove any non-prime numbers: {acc} and {n}. Return the result as a list.",
            ...     return_type=Bare(list[int]),
            ...     associative=True,
            ...     model="openai/o4-mini",
            ... )
            [2, 3, 5, 7, 11, 13, 17, 19]

            Distinguishing between leaf nodes and internal nodes in an associative reduce with [Box][semlib.box.Box]:

            >>> reviews: list[str] = [
            ...     "The instructions are a bit confusing. It took me a while to figure out how to use it.",
            ...     "It's so loud!",
            ...     "I regret buying this microwave. It's the worst appliance I've ever owned.",
            ...     "This microwave is great! It heats up food quickly and evenly.",
            ...     "This microwave is a waste of money. It doesn't work at all.",
            ...     "I hate the design of this microwave. It looks cheap and ugly.",
            ...     "The turntable is a bit small, so I can't fit larger plates in it.",
            ...     "The microwave is a bit noisy when it's running.",
            ...     "The microwave is a bit expensive compared to other models with similar features.",
            ...     "The turntable is useless, so I can't fit any plates in it.",
            ...     "I love the sleek design of this microwave. It looks great in my kitchen.",
            ... ]
            >>> def template(a: str | Box[str], b: str | Box[str]) -> str:
            ...     # leaf nodes (raw reviews)
            ...     if isinstance(a, Box) and isinstance(b, Box):
            ...         return f'''
            ... Consider the following two product reviews, and return a bulleted list
            ... summarizing any actionable product improvements that could be made based on
            ... the reviews. If there are no actionable product improvements, return an empty
            ... string.
            ...
            ... - Review 1: {a.value}
            ... - Review 2: {b.value}'''
            ...     # summaries of reviews
            ...     if not isinstance(a, Box) and not isinstance(b, Box):
            ...         return f'''
            ... Consider the following two lists of ideas for product improvements, and
            ... combine them while de-duplicating similar ideas. If there are no ideas, return
            ... an empty string.
            ...
            ... # List 1:
            ... {a}
            ...
            ... # List 2:
            ... {b}'''
            ...     # one is a summary, the other is a raw review
            ...     if isinstance(a, Box) and not isinstance(b, Box):
            ...         ideas = b
            ...         review = a.value
            ...     if not isinstance(a, Box) and isinstance(b, Box):
            ...         ideas = a
            ...         review = b.value
            ...     return f'''
            ... Consider the following list of ideas for product improvements, and a product
            ... review. Update the list of ideas based on the review, de-duplicating similar
            ... ideas. If there are no ideas, return an empty string.
            ...
            ... # List of ideas:
            ... {ideas}
            ...
            ... # Review:
            ... {review}'''
            >>> result = await session.reduce(
            ...     map(Box, reviews), template=template, associative=True
            ... )
            >>> print(result)
            - Clarify and simplify the product instructions to make them easier to understand.
            - Consider reducing the noise level of the product to make it quieter during operation.
            - Improve product reliability to ensure the microwave functions correctly for all users.
            - Increase the size or adjust the design of the turntable to accommodate larger plates.
            - Improve the design to enhance the aesthetic appeal and make it look more premium.
        """
        if initial is not None:
            return await self._reduce2(
                iterable,
                template,
                initial,
                return_type=return_type,
                model=model,
            )
        return await self._reduce1(
            iterable,
            template,
            return_type=return_type,
            associative=associative,
            model=model,
        )

    async def _reduce1(
        self,
        iterable: Iterable[Any],
        /,
        template: str | Callable[[Any, Any], str],
        *,
        return_type: Any,
        associative: bool = False,
        model: str | None = None,
    ) -> Any:
        formatter = template.format if isinstance(template, str) else template
        if not associative:
            it = iter(iterable)
            try:
                acc = next(it)
            except StopIteration:
                msg = "reduce of empty iterable with no initial value"
                raise ValueError(msg) from None
            for item in it:
                acc = await self.prompt(formatter(acc, item), model=model, return_type=return_type)
            return acc
        # associative
        items = list(iterable)
        if len(items) == 0:
            msg = "reduce of empty iterable with no initial value"
            raise ValueError(msg) from None
        if len(items) == 1:
            return items[0]
        mid = len(items) // 2
        left = items[:mid]
        right = items[mid:]
        left_acc, right_acc = await gather(
            self._reduce1(left, template, return_type=return_type, associative=True, model=model),
            self._reduce1(right, template, return_type=return_type, associative=True, model=model),
        )
        return await self.prompt(formatter(left_acc, right_acc), model=model, return_type=return_type)

    async def _reduce2(
        self,
        iterable: Iterable[Any],
        /,
        template: str | Callable[[Any, Any], str],
        initial: Any,
        *,
        return_type: Any,
        model: str | None = None,
    ) -> Any:
        acc = initial
        formatter = template.format if isinstance(template, str) else template
        for item in iterable:
            acc = await self.prompt(formatter(acc, item), model=model, return_type=return_type)
        return acc


@overload
async def reduce(
    iterable: Iterable[str],
    /,
    template: str | Callable[[str, str], str],
    *,
    associative: bool = False,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> str: ...


@overload
async def reduce[T](
    iterable: Iterable[str | T],
    /,
    template: str | Callable[[str | T, str | T], str],
    *,
    associative: bool = False,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> str | T: ...


@overload
async def reduce[T: BaseModel](
    iterable: Iterable[T],
    /,
    template: str | Callable[[T, T], str],
    *,
    return_type: type[T],
    associative: bool = False,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> T: ...


@overload
async def reduce[T](
    iterable: Iterable[T],
    /,
    template: str | Callable[[T, T], str],
    *,
    return_type: Bare[T],
    associative: bool = False,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> T: ...


@overload
async def reduce[T, U: BaseModel](
    iterable: Iterable[T],
    /,
    template: str | Callable[[U, T], str],
    initial: U,
    *,
    return_type: type[U],
    model: str | None = None,
    max_concurrency: int | None = None,
) -> U: ...


@overload
async def reduce[T, U](
    iterable: Iterable[T],
    /,
    template: str | Callable[[U, T], str],
    initial: U,
    *,
    return_type: Bare[U],
    model: str | None = None,
    max_concurrency: int | None = None,
) -> U: ...


async def reduce(
    iterable: Iterable[Any],
    /,
    template: str | Callable[[Any, Any], str],
    initial: Any = None,
    *,
    return_type: Any = None,
    associative: bool = False,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> Any:
    """Standalone version of [reduce][semlib.reduce.Reduce.reduce]."""
    reducer = Reduce(model=model, max_concurrency=max_concurrency)
    return await reducer.reduce(  # type: ignore[call-overload]
        iterable,
        template,
        initial,
        return_type=return_type,
        associative=associative,
    )


@overload
def reduce_sync(
    iterable: Iterable[str],
    /,
    template: str | Callable[[str, str], str],
    *,
    associative: bool = False,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> str: ...


@overload
def reduce_sync[T](
    iterable: Iterable[str | T],
    /,
    template: str | Callable[[str | T, str | T], str],
    *,
    associative: bool = False,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> str | T: ...


@overload
def reduce_sync[T: BaseModel](
    iterable: Iterable[T],
    /,
    template: str | Callable[[T, T], str],
    *,
    return_type: type[T],
    associative: bool = False,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> T: ...


@overload
def reduce_sync[T](
    iterable: Iterable[T],
    /,
    template: str | Callable[[T, T], str],
    *,
    return_type: Bare[T],
    associative: bool = False,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> T: ...


@overload
def reduce_sync[T, U: BaseModel](
    iterable: Iterable[T],
    /,
    template: str | Callable[[U, T], str],
    initial: U,
    *,
    return_type: type[U],
    model: str | None = None,
    max_concurrency: int | None = None,
) -> U: ...


@overload
def reduce_sync[T, U](
    iterable: Iterable[T],
    /,
    template: str | Callable[[U, T], str],
    initial: U,
    *,
    return_type: Bare[U],
    model: str | None = None,
    max_concurrency: int | None = None,
) -> U: ...


def reduce_sync(
    iterable: Iterable[Any],
    /,
    template: str | Callable[[Any, Any], str],
    initial: Any = None,
    *,
    return_type: Any = None,
    associative: bool = False,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> Any:
    """Standalone synchronous version of [reduce][semlib.reduce.Reduce.reduce]."""
    reducer = Reduce(model=model, max_concurrency=max_concurrency)
    return asyncio.run(
        reducer.reduce(  # type: ignore[call-overload]
            iterable,
            template,
            initial,
            return_type=return_type,
            associative=associative,
        )
    )



import json
import sqlite3
from abc import ABC, abstractmethod
from hashlib import sha256
from typing import cast, override

from litellm.types.utils import Message
from pydantic import BaseModel

type CacheKey[T: BaseModel] = tuple[list[Message], type[T] | None, str]


class QueryCache(ABC):
    """Abstract base class for a cache of LLM query results.

    Caches can be used with [Session][semlib.session.Session] to avoid repeating identical queries to the LLM.
    """

    @abstractmethod
    def _set[T: BaseModel](self, key: CacheKey[T], value: str) -> None: ...

    @abstractmethod
    def _get[T: BaseModel](self, key: CacheKey[T]) -> str | None: ...

    @abstractmethod
    def clear(self) -> None: ...

    @abstractmethod
    def __len__(self) -> int: ...

    def _hash_key[T: BaseModel](self, key: CacheKey[T]) -> bytes:
        messages, pydantic_model, llm_model = key
        key_components: list[str] = [llm_model]
        key_components.extend(message.to_json() for message in messages)
        if pydantic_model is not None:
            key_components.append(json.dumps(pydantic_model.model_json_schema()))
        h = sha256()
        for part in key_components:
            h.update(part.encode("utf-8"))
        return h.digest()


class InMemoryCache(QueryCache):
    """An in-memory cache of LLM query results."""

    @override
    def __init__(self) -> None:
        """Initialize an in-memory cache."""
        self._data: dict[bytes, str] = {}

    @override
    def _set[T: BaseModel](self, key: CacheKey[T], value: str) -> None:
        self._data[self._hash_key(key)] = value

    @override
    def _get[T: BaseModel](self, key: CacheKey[T]) -> str | None:
        return self._data.get(self._hash_key(key))

    @override
    def clear(self) -> None:
        self._data.clear()

    @override
    def __len__(self) -> int:
        return len(self._data)


_VERSION_KEY = "version"
_VERSION = "1"


class OnDiskCache(QueryCache):
    """A persistent on-disk cache of LLM query results, backed by SQLite."""

    @override
    def __init__(self, path: str) -> None:
        """Initialize an on-disk cache.

        Args:
            path: Path to the SQLite database file. If the file does not exist, it will be created. By convention, the
                filename should have a ".db" or ".sqlite" extension.
        """
        self._conn = sqlite3.connect(path, autocommit=True)
        self._conn.execute(
            """
            CREATE TABLE IF NOT EXISTS metadata (
                key TEXT PRIMARY KEY,
                value TEXT NOT NULL
            )
            """
        )
        cur = self._conn.execute("SELECT value FROM metadata WHERE key = ?", (_VERSION_KEY,))
        row = cur.fetchone()
        if row is None:
            self._conn.execute("INSERT INTO metadata (key, value) VALUES (?, ?)", (_VERSION_KEY, _VERSION))
        elif row[0] != _VERSION:
            msg = f"cache version mismatch: expected {_VERSION}, got {row[0]}"
            raise ValueError(msg)
        self._conn.execute(
            """
            CREATE TABLE IF NOT EXISTS data (
                key BLOB PRIMARY KEY,
                value TEXT NOT NULL
            )
            """
        )

    @override
    def _set[T: BaseModel](self, key: CacheKey[T], value: str) -> None:
        self._conn.execute(
            "INSERT OR REPLACE INTO data (key, value) VALUES (?, ?)",
            (self._hash_key(key), value),
        )

    @override
    def _get[T: BaseModel](self, key: CacheKey[T]) -> str | None:
        cur = self._conn.execute("SELECT value FROM data WHERE key = ?", (self._hash_key(key),))
        row = cur.fetchone()
        if row is None:
            return None
        return cast(str, row[0])

    @override
    def clear(self) -> None:
        self._conn.execute("DELETE FROM data")

    @override
    def __len__(self) -> int:
        cur = self._conn.execute("SELECT COUNT(*) FROM data")
        row = cur.fetchone()
        return cast(int, row[0])



import asyncio
from collections.abc import Callable, Iterable

from pydantic_core import ValidationError  # noqa: F401 # used in docstrings

from semlib.compare import Compare, Order, Task
from semlib.sort.algorithm import Algorithm, BordaCount


class Sort(Compare):
    async def sort[T](
        self,
        iterable: Iterable[T],
        /,
        *,
        by: str | None = None,
        to_str: Callable[[T], str] | None = None,
        template: str | Callable[[T, T], str] | None = None,
        task: Task | str | None = None,
        algorithm: Algorithm | None = None,
        reverse: bool = False,
        model: str | None = None,
    ) -> list[T]:
        """Sort an iterable.

        This method sorts a collection of items by using a language model to perform pairwise comparisons. The sorting
        algorithm determines which comparisons to make and how to aggregate the results into a final ranking.

        This method is analogous to Python's built-in
        [`sorted`](https://docs.python.org/3/library/functions.html#sorted) function.

        Args:
            iterable: The collection of items to sort.
            by: A criteria specifying what aspect to compare by. If this is provided, `template` cannot be
                provided.
            to_str: If specified, used to convert items to string representation. Otherewise, uses `str()` on each item.
                If this is provided, a callable template cannot be provided.
            template: A custom prompt template for comparisons. Must be either a string template with two positional
                placeholders, or a callable that takes two items and returns a formatted string. If this is provided,
                `by` cannot be provided.
            task: The type of comparison task that is being performed in `template`. This allows for writing the
                template in the most convenient way possible (e.g., in some scenarios, it's easier to specify a criteria
                for which item is lesser, and in others, it's easier to specify a criteria for which item is greater).
                If this is provided, a custom `template` must also be provided.  Defaults to
                [Task.CHOOSE_GREATER][semlib.compare.Task.CHOOSE_GREATER] if not specified.
            algorithm: The sorting algorithm to use. If not specified, defaults to
                [BordaCount][semlib.sort.algorithm.BordaCount]. Different algorithms make different tradeoffs between
                accuracy, latency, and cost. See the documentation for each algorithm for details.
            reverse: If `True`, sort in descending order. If `False`, sort in ascending order.
            model: If specified, overrides the default model for this call.

        Returns:
            A new list containing all items from the iterable in sort order.

        Raises:
            ValidationError: If parsing any LLM response fails.

        Examples:
            Basic sort:
            >>> await session.sort(["blue", "red", "green"], by="wavelength", reverse=True)
            ['red', 'green', 'blue']

            Custom template and task:
            >>> from dataclasses import dataclass
            >>> @dataclass
            ... class Person:
            ...     name: str
            ...     job: str
            >>> people = [
            ...     Person(name="Barack Obama", job="President of the United States"),
            ...     Person(name="Dalai Lama", job="Spiritual Leader of Tibet"),
            ...     Person(name="Sundar Pichai", job="CEO of Google"),
            ... ]
            >>> await session.sort(
            ...     people,
            ...     template=lambda a, b: f"Which job earns more, (a) {a.job} or (b) {b.job}?",
            ... )
            [
                Person(name='Dalai Lama', job='Spiritual Leader of Tibet'),
                Person(name='Barack Obama', job='President of the United States'),
                Person(name='Sundar Pichai', job='CEO of Google'),
            ]
        """
        algorithm = algorithm if algorithm is not None else BordaCount()

        async def comparator(a: T, b: T) -> Order:
            return await self.compare(a, b, by=by, to_str=to_str, template=template, task=task, model=model)

        return await algorithm._sort(  # noqa: SLF001
            iterable, reverse=reverse, comparator=comparator, max_concurrency=self._max_concurrency
        )


async def sort[T](
    iterable: Iterable[T],
    /,
    *,
    by: str | None = None,
    to_str: Callable[[T], str] | None = None,
    template: str | Callable[[T, T], str] | None = None,
    task: Task | str | None = None,
    algorithm: Algorithm | None = None,
    reverse: bool = False,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> list[T]:
    """Standalone version of [sort][semlib.sort.Sort.sort]."""
    sorter = Sort(
        model=model,
        max_concurrency=max_concurrency,
    )
    return await sorter.sort(
        iterable, by=by, to_str=to_str, template=template, task=task, algorithm=algorithm, reverse=reverse
    )


def sort_sync[T](
    iterable: Iterable[T],
    /,
    *,
    by: str | None = None,
    to_str: Callable[[T], str] | None = None,
    template: str | Callable[[T, T], str] | None = None,
    task: Task | str | None = None,
    algorithm: Algorithm | None = None,
    reverse: bool = False,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> list[T]:
    """Standalone synchronous version of [sort][semlib.sort.Sort.sort]."""
    sorter = Sort(
        model=model,
        max_concurrency=max_concurrency,
    )
    return asyncio.run(
        sorter.sort(iterable, by=by, to_str=to_str, template=template, task=task, algorithm=algorithm, reverse=reverse)
    )



import asyncio
from collections.abc import Callable, Iterable

from pydantic_core import ValidationError  # noqa: F401 # used in docstrings

from semlib._internal import util
from semlib.compare import Compare, Order, Task


class Extrema(Compare):
    async def _get_extreme[T](
        self,
        iterable: list[T],
        /,
        *,
        find_min: bool,
        by: str | None = None,
        to_str: Callable[[T], str] | None = None,
        template: str | Callable[[T, T], str] | None = None,
        task: Task | str | None = None,
        model: str | None = None,
    ) -> T:
        if not iterable:
            msg = "iterable argument is empty"
            raise ValueError(msg)

        async def compare(a: T, b: T) -> Order:
            return await self.compare(
                a,
                b,
                by=by,
                to_str=to_str,
                template=template,
                task=task,
                model=model,
            )

        async def rec(lst: list[T]) -> T:
            if len(lst) == 1:
                return lst[0]
            mid = len(lst) // 2
            left, right = await util.gather(rec(lst[:mid]), rec(lst[mid:]))
            comparison = await compare(left, right)
            if (find_min and comparison == Order.LESS) or (not find_min and comparison == Order.GREATER):
                return left
            if (find_min and comparison == Order.GREATER) or (not find_min and comparison == Order.LESS):
                return right
            # neither is greater; we arbitrarily choose to be left-biased
            return left

        return await rec(iterable)

    async def min[T](
        self,
        iterable: Iterable[T],
        /,
        *,
        by: str | None = None,
        to_str: Callable[[T], str] | None = None,
        template: str | Callable[[T, T], str] | None = None,
        task: Task | str | None = None,
        model: str | None = None,
    ) -> T:
        """Get the smallest item in an iterable.

        This method finds the smallest item in a collection by using a language model to perform pairwise comparisons.

        This method is analogous to Python's built-in [`min`](https://docs.python.org/3/library/functions.html#min)
        function.

        Args:
            iterable: The collection of items to search.
            by: A criteria specifying what aspect to compare by. If this is provided, `template` cannot be
                provided.
            to_str: If specified, used to convert items to string representation. Otherewise, uses `str()` on each item.
                If this is provided, a callable template cannot be provided.
            template: A custom prompt template for comparisons. Must be either a string template with two positional
                placeholders, or a callable that takes two items and returns a formatted string. If this is provided,
                `by` cannot be provided.
            task: The type of comparison task that is being performed in `template`. This allows for writing the
                template in the most convenient way possible (e.g., in some scenarios, it's easier to specify a criteria
                for which item is lesser, and in others, it's easier to specify a criteria for which item is greater).
                If this is provided, a custom `template` must also be provided.  Defaults to
                [Task.CHOOSE_GREATER][semlib.compare.Task.CHOOSE_GREATER] if not specified.
            model: If specified, overrides the default model for this call.

        Returns:
            The smallest item.

        Raises:
            ValidationError: If parsing any LLM response fails.

        Examples:
            Basic usage:
            >>> await session.min(["blue", "red", "green"], by="wavelength")
            'blue'
        """
        return await self._get_extreme(
            list(iterable),
            find_min=True,
            by=by,
            to_str=to_str,
            template=template,
            task=task,
            model=model,
        )

    async def max[T](
        self,
        iterable: Iterable[T],
        /,
        *,
        by: str | None = None,
        to_str: Callable[[T], str] | None = None,
        template: str | Callable[[T, T], str] | None = None,
        task: Task | str | None = None,
        model: str | None = None,
    ) -> T:
        """Get the largest item in an iterable.

        This method finds the largest item in a collection by using a language model to perform pairwise comparisons.

        This method is analogous to Python's built-in [`max`](https://docs.python.org/3/library/functions.html#max)
        function.

        Args:
            iterable: The collection of items to search.
            by: A criteria specifying what aspect to compare by. If this is provided, `template` cannot be
                provided.
            to_str: If specified, used to convert items to string representation. Otherewise, uses `str()` on each item.
                If this is provided, a callable template cannot be provided.
            template: A custom prompt template for comparisons. Must be either a string template with two positional
                placeholders, or a callable that takes two items and returns a formatted string. If this is provided,
                `by` cannot be provided.
            task: The type of comparison task that is being performed in `template`. This allows for writing the
                template in the most convenient way possible (e.g., in some scenarios, it's easier to specify a criteria
                for which item is lesser, and in others, it's easier to specify a criteria for which item is greater).
                If this is provided, a custom `template` must also be provided.  Defaults to
                [Task.CHOOSE_GREATER][semlib.compare.Task.CHOOSE_GREATER] if not specified.
            model: If specified, overrides the default model for this call.

        Returns:
            The largest item.

        Raises:
            ValidationError: If parsing any LLM response fails.

        Examples:
            Basic usage:
            >>> await session.max(
            ...     ["LeBron James", "Kobe Bryant", "Magic Johnson"], by="assists"
            ... )
            'Magic Johnson'
        """
        return await self._get_extreme(
            list(iterable),
            find_min=False,
            by=by,
            to_str=to_str,
            template=template,
            task=task,
            model=model,
        )


async def min[T](  # noqa: A001
    iterable: Iterable[T],
    /,
    *,
    by: str | None = None,
    to_str: Callable[[T], str] | None = None,
    template: str | Callable[[T, T], str] | None = None,
    task: Task | str | None = None,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> T:
    """Standalone version of [min][semlib.extrema.Extrema.min]."""
    extrema = Extrema(model=model, max_concurrency=max_concurrency)
    return await extrema.min(iterable, by=by, to_str=to_str, template=template, task=task)


def min_sync[T](
    iterable: Iterable[T],
    /,
    *,
    by: str | None = None,
    to_str: Callable[[T], str] | None = None,
    template: str | Callable[[T, T], str] | None = None,
    task: Task | str | None = None,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> T:
    """Standalone synchronous version of [min][semlib.extrema.Extrema.min]."""
    extrema = Extrema(model=model, max_concurrency=max_concurrency)
    return asyncio.run(extrema.min(iterable, by=by, to_str=to_str, template=template, task=task))


async def max[T](  # noqa: A001
    iterable: Iterable[T],
    /,
    *,
    by: str | None = None,
    to_str: Callable[[T], str] | None = None,
    template: str | Callable[[T, T], str] | None = None,
    task: Task | str | None = None,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> T:
    """Standalone version of [max][semlib.extrema.Extrema.max]."""
    extrema = Extrema(model=model, max_concurrency=max_concurrency)
    return await extrema.max(iterable, by=by, to_str=to_str, template=template, task=task)


def max_sync[T](
    iterable: Iterable[T],
    /,
    *,
    by: str | None = None,
    to_str: Callable[[T], str] | None = None,
    template: str | Callable[[T, T], str] | None = None,
    task: Task | str | None = None,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> T:
    """Standalone synchronous version of [max][semlib.extrema.Extrema.max]."""
    extrema = Extrema(model=model, max_concurrency=max_concurrency)
    return asyncio.run(extrema.max(iterable, by=by, to_str=to_str, template=template, task=task))



import asyncio
from collections.abc import Awaitable, Callable, Coroutine, Iterable
from typing import Any

from semlib._internal.constants import DEFAULT_MAX_CONCURRENCY


class Poison:
    pass


poison = Poison()


async def foreach[T](fn: Callable[[T], Awaitable[None]], iterable: Iterable[T], /, *, max_concurrency: int) -> None:
    q: asyncio.Queue[T | Poison] = asyncio.Queue(maxsize=max_concurrency)

    async def worker() -> None:
        while True:
            item = await q.get()
            if isinstance(item, Poison):
                q.task_done()
                break
            await fn(item)
            q.task_done()

    async def producer() -> None:
        for item in iterable:
            await q.put(item)
        for _ in range(max_concurrency):
            await q.put(poison)

    try:
        async with asyncio.TaskGroup() as tg:
            for _ in range(max_concurrency):
                tg.create_task(worker())
            tg.create_task(producer())
    except ExceptionGroup as eg:
        raise eg.exceptions[0] from eg


async def gather[T](*coros: Coroutine[Any, Any, T]) -> list[T]:
    tasks: list[asyncio.Task[T]] = [asyncio.create_task(coro) for coro in coros]
    try:
        return await asyncio.gather(*tasks)
    except Exception:
        for task in tasks:
            if not task.done():
                task.cancel()
        await asyncio.wait(tasks)
        raise


def parse_max_concurrency(max_concurrency: int | None, model: str | None) -> int:
    if max_concurrency is not None and max_concurrency <= 0:
        msg = "max_concurrency must be a positive integer or None"
        raise ValueError(msg)

    if max_concurrency is not None:
        return max_concurrency
    if model is not None and model.startswith("ollama"):
        return 1
    return DEFAULT_MAX_CONCURRENCY



import asyncio
from collections.abc import Callable, Iterable
from typing import overload

from pydantic import BaseModel
from pydantic_core import ValidationError  # noqa: F401 # used in docstrings

from semlib._internal import util
from semlib._internal.base import Base
from semlib.bare import Bare


class Map(Base):
    @overload
    async def map[T, U: BaseModel](
        self,
        iterable: Iterable[T],
        /,
        template: str | Callable[[T], str],
        *,
        return_type: type[U],
        model: str | None = None,
    ) -> list[U]: ...

    @overload
    async def map[T, U](
        self,
        iterable: Iterable[T],
        /,
        template: str | Callable[[T], str],
        *,
        return_type: Bare[U],
        model: str | None = None,
    ) -> list[U]: ...

    @overload
    async def map[T](
        self,
        iterable: Iterable[T],
        /,
        template: str | Callable[[T], str],
        *,
        return_type: None = None,
        model: str | None = None,
    ) -> list[str]: ...

    async def map[T, U: BaseModel, V](
        self,
        iterable: Iterable[T],
        /,
        template: str | Callable[[T], str],
        *,
        return_type: type[U] | Bare[V] | None = None,
        model: str | None = None,
    ) -> list[U] | list[V] | list[str]:
        """Map a prompt template over an iterable and get responses from the language model.

        This method applies a prompt template to each item in the provided iterable, sends the resulting prompts to the
        language model, and collects the responses. The responses can be returned as raw strings, parsed into
        [Pydantic](https://pydantic.dev/) models, or extracted as bare values using the [Bare][semlib.bare.Bare] marker.

        This method is analogous to Python's built-in [`map`](https://docs.python.org/3/library/functions.html#map)
        function.

        Args:
            iterable: The collection of items to map over.
            template: A prompt template to apply to each item. This can be either a string template with a single
                positional placeholder, or a callable that takes an item and returns a formatted string.
            return_type: If not specified, the responses are returned as raw strings. If a Pydantic model class is provided,
                the responses are parsed into instances of that model. If a [Bare][semlib.bare.Bare] instance is
                provided, single values of the specified type are extracted from the responses.
            model: If specified, overrides the default model for this call.

        Returns:
            A list of responses from the language model in the format specified by return_type.

        Raises:
            ValidationError: If `return_type` is a Pydantic model or a Bare type and the response cannot be
                parsed into the specified type.

        Examples:
            Basic map:
            >>> await session.map(
            ...     ["apple", "banana", "kiwi"],
            ...     template="What color is {}? Reply in a single word.",
            ... )
            ['Red.', 'Yellow.', 'Green.']

            Map with structured return type:
            >>> class Person(pydantic.BaseModel):
            ...     name: str
            ...     age: int
            >>> await session.map(
            ...     ["Barack Obama", "Angela Merkel"],
            ...     template="Who is {}?",
            ...     return_type=Person,
            ... )
            [Person(name='Barack Obama', age=62), Person(name='Angela Merkel', age=69)]

            Map with bare return type:
            >>> await session.map(
            ...     [42, 1337, 2025],
            ...     template="What are the unique prime factors of {}?",
            ...     return_type=Bare(list[int]),
            ... )
            [[2, 3, 7], [7, 191], [3, 5]]
        """

        formatter = template.format if isinstance(template, str) else template
        model = model if model is not None else self._model
        # case analysis for type checker
        if return_type is None:
            return await util.gather(*[self.prompt(formatter(item), model=model) for item in iterable])
        if isinstance(return_type, Bare):
            return await util.gather(
                *[self.prompt(formatter(item), return_type=return_type, model=model) for item in iterable]
            )
        return await util.gather(
            *[self.prompt(formatter(item), return_type=return_type, model=model) for item in iterable]
        )


@overload
async def map[T, U: BaseModel](  # noqa: A001
    iterable: Iterable[T],
    /,
    template: str | Callable[[T], str],
    *,
    return_type: type[U],
    model: str | None = None,
    max_concurrency: int | None = None,
) -> list[U]: ...


@overload
async def map[T, U](  # noqa: A001
    iterable: Iterable[T],
    /,
    template: str | Callable[[T], str],
    *,
    return_type: Bare[U],
    model: str | None = None,
    max_concurrency: int | None = None,
) -> list[U]: ...


@overload
async def map[T](  # noqa: A001
    iterable: Iterable[T],
    /,
    template: str | Callable[[T], str],
    *,
    return_type: None = None,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> list[str]: ...


async def map[T, U: BaseModel, V](  # noqa: A001
    iterable: Iterable[T],
    /,
    template: str | Callable[[T], str],
    *,
    return_type: type[U] | Bare[V] | None = None,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> list[U] | list[V] | list[str]:
    """Standalone version of [map][semlib.map.Map.map]."""
    mapper = Map(model=model, max_concurrency=max_concurrency)
    return await mapper.map(iterable, template, return_type=return_type)


@overload
def map_sync[T, U: BaseModel](
    iterable: Iterable[T],
    /,
    template: str | Callable[[T], str],
    *,
    return_type: type[U],
    model: str | None = None,
    max_concurrency: int | None = None,
) -> list[U]: ...


@overload
def map_sync[T, U](
    iterable: Iterable[T],
    /,
    template: str | Callable[[T], str],
    *,
    return_type: Bare[U],
    model: str | None = None,
    max_concurrency: int | None = None,
) -> list[U]: ...


@overload
def map_sync[T](
    iterable: Iterable[T],
    /,
    template: str | Callable[[T], str],
    *,
    return_type: None = None,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> list[str]: ...


def map_sync[T, U: BaseModel, V](
    iterable: Iterable[T],
    /,
    template: str | Callable[[T], str],
    *,
    return_type: type[U] | Bare[V] | None = None,
    model: str | None = None,
    max_concurrency: int | None = None,
) -> list[U] | list[V] | list[str]:
    """Standalone synchronous version of [map][semlib.map.Map.map]."""
    mapper = Map(model=model, max_concurrency=max_concurrency)
    return asyncio.run(mapper.map(iterable, template, return_type=return_type))  # type: ignore[return-value]



from semlib.sort.algorithm.algorithm import Algorithm
from semlib.sort.algorithm.borda_count import BordaCount
from semlib.sort.algorithm.quicksort import QuickSort

__all__ = ["Algorithm", "BordaCount", "QuickSort"]



import secrets
from collections.abc import Callable, Coroutine, Iterable
from typing import override

from semlib._internal import util
from semlib.compare import Order
from semlib.sort.algorithm.algorithm import Algorithm


class QuickSort(Algorithm):
    """Quicksort sorting algorithm.

    This algorithm uses the [Quicksort](https://en.wikipedia.org/wiki/Quicksort) method to sort items. This algorithm
    does **not** provide theoretical guarantees with noisy pairwise comparisons, but standard sorting algorithms can
    perform well in practice ([Qin et al., 2024](https://aclanthology.org/2024.findings-naacl.97/)) even with noisy
    pariwise comparisons using LLMs.

    This algorithm requires O(n log n) pairwise comparisons on average. If you want higher-quality rankings and can
    tolerate increased costs and latency, you can consider using the [BordaCount][semlib.sort.algorithm.BordaCount]
    algorithm instead.
    """

    @override
    def __init__(self, *, randomized: bool = False):
        """Initialize.

        Args:
            randomized: If `True`, uses a randomized pivot selection strategy. This can help avoid worst-case O(n^2)
                performance on certain inputs, but results may be non-deterministic. If False, always uses the first
                item as the pivot.
        """
        super().__init__()
        self._randomized = randomized

    @override
    async def _sort[T](
        self,
        iterable: Iterable[T],
        /,
        *,
        reverse: bool = False,
        comparator: Callable[[T, T], Coroutine[None, None, Order]],
        max_concurrency: int,
    ) -> list[T]:
        lst = iterable if isinstance(iterable, list) else list(iterable)

        async def quicksort(lst: list[T]) -> list[T]:
            if len(lst) <= 1:
                return lst
            pivot_index = secrets.randbelow(len(lst)) if self._randomized else 0
            pivot = lst[pivot_index]
            less = []
            greater = []
            equal = [pivot]  # to handle "neither" case
            comparisons = await util.gather(
                *(comparator(item, pivot) for i, item in enumerate(lst) if i != pivot_index)
            )
            for i, item in enumerate(lst):
                if i == pivot_index:
                    continue
                comparison = comparisons[i if i < pivot_index else i - 1]
                if comparison == Order.LESS:
                    less.append(item)
                elif comparison == Order.GREATER:
                    greater.append(item)
                else:
                    equal.append(item)
            sort_less, sort_greater = await util.gather(quicksort(less), quicksort(greater))
            return sort_less + equal + sort_greater

        sort_list = await quicksort(lst)
        return sort_list[::-1] if reverse else sort_list



import asyncio
import datetime
from collections.abc import Callable
from typing import Any
from unittest.mock import patch

import pydantic
import pytest

from semlib.bare import Bare
from semlib.map import map, map_sync
from tests.conftest import LLMMocker


def test_map_str(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "What is the number that comes after one? Reply with a single number, e.g., '5'.": "2",
            "What is the number that comes after two? Reply with a single number, e.g., '5'.": "3",
            "What is the number that comes after three? Reply with a single number, e.g., '5'.": "4",
        }
    )

    with mocker.patch_llm():
        result: list[str] = map_sync(
            ["one", "two", "three"], "What is the number that comes after {}? Reply with a single number, e.g., '5'."
        )

    assert result == ["2", "3", "4"]


def test_map_structured_output(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    class Person(pydantic.BaseModel):
        name: str
        date_of_birth: datetime.date

    mocker = llm_mocker(
        {
            "Who was the 40th president of the United States?": '{"name": "Ronald Reagan", "date_of_birth": "1911-02-06"}',
            "Who was the 41th president of the United States?": '{"name": "George H. W. Bush", "date_of_birth": "1924-06-12"}',
            "Who was the 42th president of the United States?": '{"name": "William Jefferson Clinton", "date_of_birth": "1946-08-19"}',
            "Who was the 43th president of the United States?": '{"name": "George W. Bush", "date_of_birth": "1946-07-06"}',
        }
    )

    with mocker.patch_llm():
        result: list[Person] = map_sync(
            [40, 41, 42, 43], "Who was the {}th president of the United States?", return_type=Person
        )

    assert result == [
        Person(name="Ronald Reagan", date_of_birth=datetime.date(1911, 2, 6)),
        Person(name="George H. W. Bush", date_of_birth=datetime.date(1924, 6, 12)),
        Person(name="William Jefferson Clinton", date_of_birth=datetime.date(1946, 8, 19)),
        Person(name="George W. Bush", date_of_birth=datetime.date(1946, 7, 6)),
    ]


def test_map_callable_formatter(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    class Car(pydantic.BaseModel):
        make: str
        model: str
        year: int

    def template(car: Car) -> str:
        return f"What is the horsepower of a {car.year} {car.make} {car.model}?"

    cars: list[Car] = [
        Car(make="Toyota", model="Camry", year=2020),
        Car(make="Honda", model="Pilot", year=2019),
        Car(make="Ford", model="F-450", year=2021),
    ]

    class Engine(pydantic.BaseModel):
        horsepower: int

    mocker = llm_mocker(
        {
            "What is the horsepower of a 2020 Toyota Camry?": '{"horsepower": 203}',
            "What is the horsepower of a 2019 Honda Pilot?": '{"horsepower": 280}',
            "What is the horsepower of a 2021 Ford F-450?": '{"horsepower": 475}',
        }
    )

    with mocker.patch_llm():
        result: list[Engine] = map_sync(cars, template, return_type=Engine)

    assert result == [Engine(horsepower=203), Engine(horsepower=280), Engine(horsepower=475)]


@pytest.mark.asyncio
async def test_map_async(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "What is 1 + 1?": "2",
            "What is 1 + 3?": "4",
            "What is 1 + 5?": "6",
        }
    )

    with mocker.patch_llm():
        result: list[str] = await map([1, 3, 5], "What is 1 + {:d}?")

    assert result == ["2", "4", "6"]


def test_map_raises(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "What is the number that comes after one?": "not json",
            "What is the number that comes after two?": '{"value": 3}',
            "What is the number that comes after three?": '{"value": 4}',
        }
    )

    class Number(pydantic.BaseModel):
        value: int

    with mocker.patch_llm(), pytest.raises(pydantic.ValidationError):
        map_sync(["one", "two", "three"], "What is the number that comes after {}?", return_type=Number)


def test_map_bare(llm_mocker: Callable[[dict[str, str]], LLMMocker]) -> None:
    mocker = llm_mocker(
        {
            "What is the number that comes after one?": '{"value": 2}',
            "What is the number that comes after two?": '{"value": 3}',
            "What is the number that comes after three?": '{"value": 4}',
        }
    )

    with mocker.patch_llm():
        result = map_sync(["one", "two", "three"], "What is the number that comes after {}?", return_type=Bare(int))

    assert result == [2, 3, 4]


@pytest.mark.asyncio
async def test_map_task_cleanup_on_exception() -> None:
    task_cancelled = asyncio.Event()

    async def mock_acompletion(*_args: Any, **kwargs: Any) -> Any:
        content = kwargs.get("messages", [{}])[0].get("content", "")

        if "item_1" in content:
            # first task fails after a short delay
            await asyncio.sleep(0.1)
            msg = "Mock API error"
            raise ValueError(msg)

        # "item_2" case: other task takes a long time to complete
        try:
            q: asyncio.Queue[None] = asyncio.Queue()
            await q.get()
        except asyncio.CancelledError:
            task_cancelled.set()
            raise

    with patch("litellm.acompletion", side_effect=mock_acompletion), patch("litellm.completion_cost", return_value=0.0):
        with pytest.raises(ValueError, match="Mock API error"):
            await map(["item_1", "item_2"], template="{}")

        await task_cancelled.wait()  # this will hang if the task wasn't cancelled
